from __future__ import annotations

import itertools

from mpi4py import MPI

import dolfinx
import numpy as np
import pytest

import io4dolfinx

root = 0
dtypes: list["str"] = ["float64", "float32"]  # Mesh geometry dtypes
write_comm: list[MPI.Intracomm] = [
    MPI.COMM_SELF,
    MPI.COMM_WORLD,
]  # Communicators for creating mesh
read_modes: list[dolfinx.mesh.GhostMode] = [
    dolfinx.mesh.GhostMode.none,
    dolfinx.mesh.GhostMode.shared_facet,
]
# Cell types of different dimensions
two_dimensional_cell_types: list[dolfinx.mesh.CellType] = [
    dolfinx.mesh.CellType.triangle,
    dolfinx.mesh.CellType.quadrilateral,
]
three_dimensional_cell_types: list[dolfinx.mesh.CellType] = [
    dolfinx.mesh.CellType.tetrahedron,
    dolfinx.mesh.CellType.hexahedron,
]

one_dim_combinations = itertools.product(dtypes, write_comm)
two_dim_combinations = itertools.product(dtypes, two_dimensional_cell_types, write_comm)
three_dim_combinations = itertools.product(dtypes, three_dimensional_cell_types, write_comm)


@pytest.fixture(params=one_dim_combinations, scope="module")
def mesh_1D(request):
    dtype, write_comm = request.param
    mesh = dolfinx.mesh.create_unit_interval(write_comm, 8, dtype=np.dtype(dtype))
    return mesh


@pytest.fixture(params=two_dim_combinations, scope="module")
def mesh_2D(request):
    dtype, cell_type, write_comm = request.param
    mesh = dolfinx.mesh.create_unit_square(
        write_comm, 10, 7, cell_type=cell_type, dtype=np.dtype(dtype)
    )
    return mesh


@pytest.fixture(params=three_dim_combinations, scope="module")
def mesh_3D(request):
    dtype, cell_type, write_comm = request.param
    mesh = dolfinx.mesh.create_unit_cube(
        write_comm, 5, 7, 3, cell_type=cell_type, dtype=np.dtype(dtype)
    )
    return mesh


@pytest.mark.parametrize("read_mode", read_modes)
@pytest.mark.parametrize("read_comm", [MPI.COMM_SELF, MPI.COMM_WORLD])
def test_checkpointing_meshtags_1D(
    mesh_1D, read_comm, read_mode, tmp_path, backend, generate_reference_map
):
    mesh = mesh_1D

    # Write unique mesh file for each combination of MPI communicator and dtype
    hash = f"{mesh.comm.size}_{mesh.geometry.x.dtype}"
    fname = MPI.COMM_WORLD.bcast(tmp_path, root=0)
    suffix = ".bp" if backend == "adios" else ".h5"
    filename = (fname / f"{backend}_meshtags_1D_{hash}").with_suffix(suffix)

    # If mesh communicator is more than a self communicator or serial write on all processes.
    # If serial or self communicator, only write on root rank
    if mesh.comm.size != 1:
        io4dolfinx.write_mesh(filename, mesh, backend=backend)
    else:
        if MPI.COMM_WORLD.rank == root:
            io4dolfinx.write_mesh(filename, mesh, backend=backend)

    # Create meshtags labeling each entity (of each co-dimension) with a
    # unique number (their initial global index).
    org_maps = []
    for dim in range(mesh.topology.dim + 1):
        mesh.topology.create_connectivity(dim, mesh.topology.dim)
        e_map = mesh.topology.index_map(dim)
        num_entities_local = e_map.size_local
        entities = np.arange(num_entities_local, dtype=np.int32)
        ft = dolfinx.mesh.meshtags(mesh, dim, entities, e_map.local_range[0] + entities)
        ft.name = f"entity_{dim}"

        # If parallel write on all processes, else write on root rank
        if mesh.comm.size != 1:
            io4dolfinx.write_meshtags(filename, mesh, ft, backend=backend)
            # Create map from mesh tag value to its corresponding index and midpoint
            org_map = generate_reference_map(mesh, ft, mesh.comm, root)
            org_maps.append(org_map)
        else:
            if MPI.COMM_WORLD.rank == root:
                io4dolfinx.write_meshtags(filename, mesh, ft, backend=backend)
                # Create map from mesh tag value to its corresponding index and midpoint
                org_map = generate_reference_map(mesh, ft, MPI.COMM_SELF, root)
                org_maps.append(org_map)
        del ft
    del mesh
    MPI.COMM_WORLD.Barrier()
    # Read mesh on testing communicator
    new_mesh = io4dolfinx.read_mesh(filename, read_comm, ghost_mode=read_mode, backend=backend)
    for dim in range(new_mesh.topology.dim + 1):
        # Read meshtags on all processes if testing communicator has multiple ranks
        # else read on root 0
        if read_comm.size != 1:
            new_ft = io4dolfinx.read_meshtags(
                filename,
                new_mesh,
                meshtag_name=f"entity_{dim}",
                backend=backend,
            )
            # Generate meshtags map from mesh tag value to its corresponding index and midpoint
            # and gather on root process
            read_map = generate_reference_map(new_mesh, new_ft, new_mesh.comm, root)
        else:
            if MPI.COMM_WORLD.rank == root:
                new_ft = io4dolfinx.read_meshtags(
                    filename,
                    new_mesh,
                    meshtag_name=f"entity_{dim}",
                    backend=backend,
                )
                read_map = generate_reference_map(new_mesh, new_ft, read_comm, root)

        # On root process, check that midpoints are the same for each value in the meshtag
        if MPI.COMM_WORLD.rank == root:
            org_map = org_maps[dim]
            assert len(org_map) == len(read_map)
            for value, (_, midpoint) in org_map.items():
                _, read_midpoint = read_map[value]
                np.testing.assert_allclose(read_midpoint, midpoint)


@pytest.mark.parametrize("read_mode", read_modes)
@pytest.mark.parametrize("read_comm", [MPI.COMM_SELF, MPI.COMM_WORLD])
def test_checkpointing_meshtags_2D(
    mesh_2D, read_comm, read_mode, tmp_path, backend, generate_reference_map
):
    mesh = mesh_2D
    hash = f"{mesh.comm.size}_{mesh.topology.cell_name()}_{mesh.geometry.x.dtype}"
    fname = MPI.COMM_WORLD.bcast(tmp_path, root=0)
    filename = fname / f"meshtags_1D_{hash}.bp"

    if mesh.comm.size != 1:
        io4dolfinx.write_mesh(filename, mesh, backend=backend)
    else:
        if MPI.COMM_WORLD.rank == root:
            io4dolfinx.write_mesh(filename, mesh, backend=backend)

    org_maps = []
    for dim in range(mesh.topology.dim + 1):
        mesh.topology.create_connectivity(dim, mesh.topology.dim)
        e_map = mesh.topology.index_map(dim)
        num_entities_local = e_map.size_local
        entities = np.arange(num_entities_local, dtype=np.int32)
        ft = dolfinx.mesh.meshtags(mesh, dim, entities, e_map.local_range[0] + entities)
        ft.name = f"entity_{dim}"
        if mesh.comm.size != 1:
            io4dolfinx.write_meshtags(filename, mesh, ft, backend=backend)
            org_map = generate_reference_map(mesh, ft, mesh.comm, root)
            org_maps.append(org_map)
        else:
            if MPI.COMM_WORLD.rank == root:
                io4dolfinx.write_meshtags(filename, mesh, ft, backend=backend)
                org_map = generate_reference_map(mesh, ft, MPI.COMM_SELF, root)
                org_maps.append(org_map)
        del ft
    del mesh
    MPI.COMM_WORLD.Barrier()
    new_mesh = io4dolfinx.read_mesh(filename, read_comm, ghost_mode=read_mode, backend=backend)
    for dim in range(new_mesh.topology.dim + 1):
        if read_comm.size != 1:
            new_ft = io4dolfinx.read_meshtags(
                filename,
                new_mesh,
                meshtag_name=f"entity_{dim}",
                backend=backend,
            )
            read_map = generate_reference_map(new_mesh, new_ft, new_mesh.comm, root)
        else:
            if MPI.COMM_WORLD.rank == root:
                new_ft = io4dolfinx.read_meshtags(
                    filename,
                    new_mesh,
                    meshtag_name=f"entity_{dim}",
                    backend=backend,
                )
                read_map = generate_reference_map(new_mesh, new_ft, read_comm, root)

        if MPI.COMM_WORLD.rank == root:
            org_map = org_maps[dim]
            assert len(org_map) == len(read_map)
            for value, (_, midpoint) in org_map.items():
                _, read_midpoint = read_map[value]
                np.testing.assert_allclose(read_midpoint, midpoint)


@pytest.mark.parametrize("read_mode", read_modes)
@pytest.mark.parametrize("read_comm", [MPI.COMM_SELF, MPI.COMM_WORLD])
def test_checkpointing_meshtags_3D(
    mesh_3D, read_comm, read_mode, tmp_path, backend, generate_reference_map
):
    mesh = mesh_3D
    hash = f"{mesh.comm.size}_{mesh.topology.cell_name()}_{mesh.geometry.x.dtype}"
    fname = MPI.COMM_WORLD.bcast(tmp_path, root=0)
    filename = fname / f"meshtags_1D_{hash}.bp"
    if mesh.comm.size != 1:
        io4dolfinx.write_mesh(filename, mesh, backend=backend)
    else:
        if MPI.COMM_WORLD.rank == root:
            io4dolfinx.write_mesh(filename, mesh, backend=backend)

    org_maps = []
    for dim in range(mesh.topology.dim + 1):
        mesh.topology.create_connectivity(dim, mesh.topology.dim)
        e_map = mesh.topology.index_map(dim)
        num_entities_local = e_map.size_local
        entities = np.arange(num_entities_local, dtype=np.int32)
        ft = dolfinx.mesh.meshtags(mesh, dim, entities, e_map.local_range[0] + entities)
        ft.name = f"entity_{dim}"

        if mesh.comm.size != 1:
            io4dolfinx.write_meshtags(filename, mesh, ft, backend=backend)
            org_map = generate_reference_map(mesh, ft, mesh.comm, root)
            org_maps.append(org_map)
        else:
            if MPI.COMM_WORLD.rank == root:
                io4dolfinx.write_meshtags(filename, mesh, ft, backend=backend)
                org_map = generate_reference_map(mesh, ft, MPI.COMM_SELF, root)
                org_maps.append(org_map)
        del ft
    del mesh

    MPI.COMM_WORLD.Barrier()
    new_mesh = io4dolfinx.read_mesh(filename, read_comm, ghost_mode=read_mode, backend=backend)
    for dim in range(new_mesh.topology.dim + 1):
        if read_comm.size != 1:
            new_ft = io4dolfinx.read_meshtags(
                filename,
                new_mesh,
                meshtag_name=f"entity_{dim}",
                backend=backend,
            )
            read_map = generate_reference_map(new_mesh, new_ft, new_mesh.comm, root)
        else:
            if MPI.COMM_WORLD.rank == root:
                new_ft = io4dolfinx.read_meshtags(
                    filename,
                    new_mesh,
                    meshtag_name=f"entity_{dim}",
                    backend=backend,
                )
                read_map = generate_reference_map(new_mesh, new_ft, MPI.COMM_SELF, root)
        if MPI.COMM_WORLD.rank == root:
            org_map = org_maps[dim]
            assert len(org_map) == len(read_map)
            for value, (_, midpoint) in org_map.items():
                _, read_midpoint = read_map[value]
                np.testing.assert_allclose(read_midpoint, midpoint)



from mpi4py import MPI

import dolfinx
import numpy as np
import pytest
import ufl

from io4dolfinx import reconstruct_mesh


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
@pytest.mark.parametrize("degree", [2, 3])
@pytest.mark.parametrize("R", [0.1, 1, 10])
def test_curve_mesh(degree, dtype, R):
    N = 8
    mesh = dolfinx.mesh.create_rectangle(
        MPI.COMM_WORLD,
        [[-1, -1], [1, 1]],
        [N, N],
        diagonal=dolfinx.mesh.DiagonalType.crossed,
        dtype=dtype,
    )
    org_area = dolfinx.fem.form(1 * ufl.dx(domain=mesh), dtype=dtype)

    curved_mesh = reconstruct_mesh(mesh, degree)

    def transform(x):
        u = R * x[0] * np.sqrt(1 - (x[1] ** 2 / (2)))
        v = R * x[1] * np.sqrt(1 - (x[0] ** 2 / (2)))
        return np.asarray([u, v])

    curved_mesh.geometry.x[:, : curved_mesh.geometry.dim] = transform(curved_mesh.geometry.x.T).T

    area = dolfinx.fem.form(1 * ufl.dx(domain=curved_mesh), dtype=dtype)
    circumference = dolfinx.fem.form(1 * ufl.ds(domain=curved_mesh), dtype=dtype)

    computed_area = curved_mesh.comm.allreduce(dolfinx.fem.assemble_scalar(area), op=MPI.SUM)
    computed_circumference = curved_mesh.comm.allreduce(
        dolfinx.fem.assemble_scalar(circumference), op=MPI.SUM
    )

    tol = 10 * np.finfo(dtype).eps
    assert np.isclose(computed_area, np.pi * R**2, atol=tol)
    assert np.isclose(computed_circumference, 2 * np.pi * R, atol=tol)

    linear_mesh = reconstruct_mesh(curved_mesh, 1)
    linear_area = dolfinx.fem.form(1 * ufl.dx(domain=linear_mesh), dtype=dtype)

    recovered_area = linear_mesh.comm.allreduce(
        dolfinx.fem.assemble_scalar(linear_area), op=MPI.SUM
    )

    # Curve original mesh
    mesh.geometry.x[:, : mesh.geometry.dim] = transform(mesh.geometry.x.T).T
    ref_area = mesh.comm.allreduce(dolfinx.fem.assemble_scalar(org_area), op=MPI.SUM)
    assert np.isclose(recovered_area, ref_area, atol=tol)



# Copyright (C) 2023 Jørgen Schartum Dokken
#
# This file is part of io4dolfinx
#
# SPDX-License-Identifier:    MIT

from __future__ import annotations

import pathlib
import typing
from pathlib import Path
from typing import Any

from mpi4py import MPI

import basix
import dolfinx
import numpy as np
import numpy.typing as npt
import ufl

from .backends import ReadMode, get_backend
from .comm_helpers import send_dofs_and_recv_values
from .utils import (
    check_file_exists,
    compute_dofmap_pos,
    compute_insert_position,
    compute_local_range,
    index_owner,
)

__all__ = ["read_mesh_from_legacy_h5", "read_function_from_legacy_h5", "read_point_data"]


def map_dofmap(dofmap: dolfinx.graph.AdjacencyList, bs: int) -> npt.NDArray[np.int64]:
    """
    Map xxxyyyzzz to xyzxyz
    """

    in_dofmap = dofmap.array
    in_offsets = dofmap.offsets

    mapped_dofmap = np.empty_like(in_dofmap)
    for i in range(len(in_offsets) - 1):
        pos_begin, pos_end = (
            in_offsets[i] - in_offsets[0],
            in_offsets[i + 1] - in_offsets[0],
        )
        dofs_i = in_dofmap[pos_begin:pos_end]
        assert (pos_end - pos_begin) % bs == 0
        num_dofs_local = int((pos_end - pos_begin) // bs)
        for k in range(bs):
            for j in range(num_dofs_local):
                mapped_dofmap[int(pos_begin + j * bs + k)] = dofs_i[int(num_dofs_local * k + j)]
    return mapped_dofmap.astype(np.int64)


def send_cells_and_receive_dofmap_index(
    filename: pathlib.Path,
    comm: MPI.Intracomm,
    source_ranks: npt.NDArray[np.int32],
    dest_ranks: npt.NDArray[np.int32],
    dest_size: npt.NDArray[np.int32],
    output_owners: npt.NDArray[np.int32],
    input_cells: npt.NDArray[np.int64],
    dofmap_pos: npt.NDArray[np.int32],
    num_cells_global: np.int64,
    dofmap_path: str,
    xdofmap_path: str,
    bs: int,
    backend: str,
) -> npt.NDArray[np.int64]:
    """
    Given a set of positions in input dofmap, give the global input index of this dofmap entry
    in input file.
    """
    check_file_exists(filename)

    recv_size = np.zeros(len(source_ranks), dtype=np.int32)
    mesh_to_data_comm = comm.Create_dist_graph_adjacent(
        source_ranks.tolist(), dest_ranks.tolist(), reorder=False
    )
    # Send sizes to create data structures for receiving from NeighAlltoAllv
    mesh_to_data_comm.Neighbor_alltoall(dest_size, recv_size)

    # Sort output for sending and fill send data
    out_cells = np.zeros(len(output_owners), dtype=np.int64)
    out_pos = np.zeros(len(output_owners), dtype=np.int32)
    proc_to_dof = np.zeros_like(input_cells, dtype=np.int32)
    insertion_array = compute_insert_position(output_owners, dest_ranks, dest_size)
    out_cells[insertion_array] = input_cells
    out_pos[insertion_array] = dofmap_pos
    proc_to_dof[insertion_array] = np.arange(len(input_cells), dtype=np.int32)
    del insertion_array

    # Prepare data-structures for receiving
    total_incoming = sum(recv_size)
    inc_cells = np.zeros(total_incoming, dtype=np.int64)
    inc_pos = np.zeros(total_incoming, dtype=np.intc)

    # Send data
    s_msg = [out_cells, dest_size, MPI.INT64_T]
    r_msg = [inc_cells, recv_size, MPI.INT64_T]
    mesh_to_data_comm.Neighbor_alltoallv(s_msg, r_msg)

    s_msg = [out_pos, dest_size, MPI.INT32_T]
    r_msg = [inc_pos, recv_size, MPI.INT32_T]
    mesh_to_data_comm.Neighbor_alltoallv(s_msg, r_msg)
    mesh_to_data_comm.Free()

    backend_cls = get_backend(backend)
    # Read dofmap from file
    backend_args = {"dofmap": dofmap_path, "offsets": xdofmap_path}
    if backend == "adios2":
        backend_args.update({"engine": "HDF5"})
    input_dofs = backend_cls.read_dofmap(filename, comm, name="", backend_args=backend_args)
    # Map to xyz
    mapped_dofmap = map_dofmap(input_dofs, bs).astype(np.int64)

    # Extract dofmap data
    local_cell_range = compute_local_range(comm, num_cells_global)
    input_cell_positions = inc_cells - local_cell_range[0]
    in_offsets = input_dofs.offsets
    read_pos = (in_offsets[input_cell_positions] + inc_pos - in_offsets[0]).astype(np.int32)
    input_dofs = mapped_dofmap[read_pos]
    del input_cell_positions, read_pos

    # Send input dofs back to owning process
    data_to_mesh_comm = comm.Create_dist_graph_adjacent(
        dest_ranks.tolist(), source_ranks.tolist(), reorder=False
    )

    incoming_global_dofs = np.zeros(sum(dest_size), dtype=np.int64)
    s_msg = [input_dofs, recv_size, MPI.INT64_T]
    r_msg = [incoming_global_dofs, dest_size, MPI.INT64_T]
    data_to_mesh_comm.Neighbor_alltoallv(s_msg, r_msg)

    # Sort incoming global dofs as they were inputted
    sorted_global_dofs = np.zeros_like(incoming_global_dofs, dtype=np.int64)
    assert len(incoming_global_dofs) == len(input_cells)
    sorted_global_dofs[proc_to_dof] = incoming_global_dofs
    data_to_mesh_comm.Free()
    return sorted_global_dofs


def read_mesh_from_legacy_h5(
    filename: pathlib.Path,
    comm: MPI.Intracomm,
    group: str,
    cell_type: str = "tetrahedron",
    backend: str = "adios2",
    max_facet_to_cell_links: int = 2,
) -> dolfinx.mesh.Mesh:
    """
    Read mesh from `h5`-file generated by legacy DOLFIN `HDF5File.write` or `XDMF.write_checkpoint`.

    Args:
        comm: MPI communicator to distribute mesh over
        filename: Path to `h5` or `xdmf` file
        group: Name of mesh in `h5`-file
        cell_type: What type of cell type, by default tetrahedron.
        backend: The IO backend to use when reading the mesh (must
            support legacy mesh reading, e.g., "adios2").
        max_facet_to_cell_links: Maximum number of cells a facet
            can be connected to.
    """
    # Make sure we use the HDF5File and check that the file is present
    check_file_exists(filename)

    backend_cls = get_backend(backend)
    mesh_topology, mesh_geometry, ct = backend_cls.read_legacy_mesh(filename, comm, group)
    if ct is not None:
        cell_type = ct
    # Create DOLFINx mesh
    element = basix.ufl.element(
        basix.ElementFamily.P,
        cell_type,
        1,
        basix.LagrangeVariant.equispaced,
        shape=(mesh_geometry.shape[1],),
    )
    domain = ufl.Mesh(element)

    try:
        return dolfinx.mesh.create_mesh(
            comm=MPI.COMM_WORLD,
            cells=mesh_topology,
            x=mesh_geometry,
            e=domain,
            partitioner=None,
            max_facet_to_cell_links=max_facet_to_cell_links,
        )
    except TypeError:
        return dolfinx.mesh.create_mesh(
            comm=MPI.COMM_WORLD,
            cells=mesh_topology,
            x=mesh_geometry,
            e=domain,
            partitioner=None,
        )

    # Should change to the commented code below when we require python
    # minimum version to be >=3.12 see https://github.com/python/cpython/pull/116198
    # import inspect
    # sig = inspect.signature(dolfinx.mesh.create_mesh)
    # kwargs: dict[str, int] = {}
    # if "max_facet_to_cell_links" in list(sig.parameters.keys()):
    #     kwargs["max_facet_to_cell_links"] = max_facet_to_cell_links

    # return dolfinx.mesh.create_mesh(
    #     comm=MPI.COMM_WORLD,
    #     cells=mesh_topology,
    #     x=mesh_geometry,
    #     e=domain,
    #     partitioner=None,
    #     **kwargs,
    # )


def read_function_from_legacy_h5(
    filename: pathlib.Path,
    comm: MPI.Intracomm,
    u: dolfinx.fem.Function,
    group: str = "mesh",
    step: typing.Optional[int] = None,
    backend: str = "adios2",
):
    """
    Read function from a `h5`-file generated by legacy DOLFIN `HDF5File.write`
    or `XDMF.write_checkpoint`.

    Args:
        comm : MPI communicator to distribute mesh over
        filename : Path to `h5` or `xdmf` file
        u : The function used to stored the read values
        group : Group within the `h5` file where the function is stored, by default "mesh"
        step : The time step used when saving the checkpoint. If not provided it will assume that
            the function is saved as a regular function (i.e with `HDF5File.write`)
        backend: The IO backend
    """

    # Make sure we use the HDF5File and check that the file is present
    filename = pathlib.Path(filename).with_suffix(".h5")
    if not filename.is_file():
        raise FileNotFoundError(f"File {filename} does not exist")

    V = u.function_space
    mesh = u.function_space.mesh
    if u.function_space.element.needs_dof_transformations:
        raise RuntimeError(
            "Function-spaces requiring dof permutations are not compatible with legacy data"
        )
    # ----------------------Step 1---------------------------------
    # Compute index of input cells, and position in input dofmap
    local_cells, dof_pos = compute_dofmap_pos(u.function_space)
    input_cells = mesh.topology.original_cell_index[local_cells]

    # Compute mesh->input communicator
    # 1.1 Compute mesh->input communicator
    num_cells_global = mesh.topology.index_map(mesh.topology.dim).size_global
    backend_cls = get_backend(backend)
    owners: npt.NDArray[np.int32]
    if backend_cls.read_mode == ReadMode.serial:
        owners = np.zeros(len(input_cells), dtype=np.int32)
    elif backend_cls.read_mode == ReadMode.parallel:
        owners = index_owner(V.mesh.comm, input_cells, num_cells_global)
    else:
        raise NotImplementedError(f"{backend_cls.read_mode} not implemented")

    unique_owners, owner_count = np.unique(owners, return_counts=True)
    # FIXME: In C++ use NBX to find neighbourhood
    _tmp_comm = mesh.comm.Create_dist_graph(
        [mesh.comm.rank], [len(unique_owners)], unique_owners, reorder=False
    )
    source, dest, _ = _tmp_comm.Get_dist_neighbors()
    _tmp_comm.Free()
    # Strip out any /
    group = group.strip("/")
    if step is not None:
        group = f"{group}/{group}_{step}"
        vector_group = "vector"
    else:
        vector_group = "vector_0"

    # ----------------------Step 2--------------------------------
    # Get global dofmap indices from input process
    bs = V.dofmap.bs
    num_cells_global = mesh.topology.index_map(mesh.topology.dim).size_global
    dofmap_indices = send_cells_and_receive_dofmap_index(
        filename,
        comm,
        np.asarray(source, dtype=np.int32),
        np.asarray(dest, dtype=np.int32),
        owner_count.astype(np.int32),
        owners,
        input_cells,
        dof_pos,
        num_cells_global,
        f"/{group}/cell_dofs",
        f"/{group}/x_cell_dofs",
        bs,
        backend=backend,
    )

    # ----------------------Step 3---------------------------------
    dof_owner: npt.NDArray[np.int32]
    if backend_cls.read_mode == ReadMode.serial:
        dof_owner = np.zeros(len(dofmap_indices), dtype=np.int32)
    elif backend_cls.read_mode == ReadMode.parallel:
        # Compute owner of global dof on distributed input data
        num_dof_global = V.dofmap.index_map_bs * V.dofmap.index_map.size_global
        dof_owner = index_owner(comm=mesh.comm, indices=dofmap_indices, N=num_dof_global)
    else:
        raise NotImplementedError(f"{backend_cls.read_mode} not implemented")

    # Create MPI neigh comm to owner.
    # NOTE: USE NBX in C++

    # Read input data
    local_array, starting_pos = backend_cls.read_hdf5_array(
        comm, filename, f"/{group}/{vector_group}", backend_args=None
    )

    # Send global dof indices to correct input process, and receive value of given dof
    local_values = send_dofs_and_recv_values(
        dofmap_indices, dof_owner, comm, local_array, starting_pos
    )

    # ----------------------Step 4---------------------------------
    # Populate local part of array and scatter forward
    u.x.array[: len(local_values)] = local_values
    u.x.scatter_forward()


def create_geometry_function_space(mesh: dolfinx.mesh.Mesh, N: int) -> dolfinx.fem.FunctionSpace:
    """Reconstruct a vector space with the N components using the geometry dofmap to ensure
    a 1-1 mapping between mesh nodes and DOFs."""
    geom_imap = mesh.geometry.index_map()
    geom_dofmap = mesh.geometry.dofmap
    ufl_domain = mesh.ufl_domain()
    assert ufl_domain is not None
    sub_el = ufl_domain.ufl_coordinate_element().sub_elements[0]
    adj_list = dolfinx.cpp.graph.AdjacencyList_int32(geom_dofmap)

    value_shape: tuple[int, ...]
    if N == 1:
        ufl_el = sub_el
        value_shape = ()
    else:
        ufl_el = basix.ufl.blocked_element(sub_el, shape=(N,))
        value_shape = (N,)

    if ufl_el.dtype == np.float32:
        _fe_constructor = dolfinx.cpp.fem.FiniteElement_float32
        _fem_constructor = dolfinx.cpp.fem.FunctionSpace_float32
    elif ufl_el.dtype == np.float64:
        _fe_constructor = dolfinx.cpp.fem.FiniteElement_float64
        _fem_constructor = dolfinx.cpp.fem.FunctionSpace_float64
    else:
        raise RuntimeError(f"Unsupported type {ufl_el.dtype}")
    try:
        cpp_el = _fe_constructor(ufl_el.basix_element._e, block_shape=value_shape, symmetric=False)
    except TypeError:
        cpp_el = _fe_constructor(ufl_el.basix_element._e, block_size=N, symmetric=False)
    dof_layout = dolfinx.cpp.fem.create_element_dof_layout(cpp_el, [])
    cpp_dofmap = dolfinx.cpp.fem.DofMap(dof_layout, geom_imap, N, adj_list, N)

    # Create function space
    try:
        cpp_space = _fem_constructor(mesh._cpp_object, cpp_el, cpp_dofmap)
    except TypeError:
        cpp_space = _fem_constructor(mesh._cpp_object, cpp_el, cpp_dofmap, value_shape=value_shape)

    return dolfinx.fem.FunctionSpace(mesh, ufl_el, cpp_space)


def read_point_data(
    filename: Path | str,
    name: str,
    mesh: dolfinx.mesh.Mesh,
    time: float | None = None,
    backend_args: dict[str, Any] | None = None,
    backend: str = "xdmf",
) -> dolfinx.fem.Function:
    """Read data from the nodes of a mesh.

    Note:
        Backend has to implement {py:class}`io4dolfinx.backends.read_cell_data`.

    Args:
        filename: Path to file
        name: Name of point data
        mesh: The corresponding :py:class:`dolfinx.mesh.Mesh`.
        time: Time-step to read from.

    Returns:
        A function in the space equivalent to the mesh
        coordinate element (up to shape).
    """

    backend_cls = get_backend(backend)
    dataset, local_range_start = backend_cls.read_point_data(
        filename=filename, name=name, comm=mesh.comm, time=time, backend_args=backend_args
    )

    num_components = dataset.shape[1]

    # Create appropriate function space (based on coordinate map)
    V = create_geometry_function_space(mesh, num_components)
    uh = dolfinx.fem.Function(V, name=name, dtype=dataset.dtype)
    # Assume that mesh is first order for now
    x_dofmap = mesh.geometry.dofmap
    igi = np.array(mesh.geometry.input_global_indices, dtype=np.int64)

    # This is dependent on how the data is read in. If distributed equally this is correct
    global_geom_input = igi[x_dofmap]

    if backend_cls.read_mode == ReadMode.parallel:
        num_nodes_global = mesh.geometry.index_map().size_global
        global_geom_owner = index_owner(mesh.comm, global_geom_input.reshape(-1), num_nodes_global)
    elif backend_cls.read_mode == ReadMode.serial:
        # This is correct if everything is read in on rank 0
        global_geom_owner = np.zeros(len(global_geom_input.flatten()), dtype=np.int32)
    else:
        raise NotImplementedError(f"{backend_cls.read_mode} not implemented")

    for i in range(num_components):
        arr_i = send_dofs_and_recv_values(
            global_geom_input.reshape(-1),
            global_geom_owner,
            mesh.comm,
            dataset[:, i],
            local_range_start,
        )
        dof_pos = x_dofmap.reshape(-1) * num_components + i
        uh.x.array[dof_pos] = arr_i
    uh.x.scatter_forward()
    return uh


def read_cell_data(
    filename: Path | str,
    name: str,
    mesh: dolfinx.mesh.Mesh,
    time: float | None = None,
    backend_args: dict[str, Any] | None = None,
    backend: str = "xdmf",
) -> dolfinx.fem.Function:
    """Read data from the nodes of a mesh.

    Note:
        Backend has to implement {py:class}`io4dolfinx.backends.read_cell_data`.

    Args:
        filename: Path to file
        name: Name of point data
        mesh: The corresponding :py:class:`dolfinx.mesh.Mesh`.
        time: Time-step to read from.

    Returns:
        A function in a DG-0 space on the mesh. The cells not found in input is set to zero.
    """

    backend_cls = get_backend(backend)

    topology, dofs = backend_cls.read_cell_data(
        filename=filename, name=name, comm=mesh.comm, time=time, backend_args=backend_args
    )
    num_components = dofs.shape[1]
    shape: tuple[int, ...]
    if num_components == 1:
        shape = ()
    else:
        shape = (num_components,)
    V = dolfinx.fem.functionspace(mesh, ("DG", 0, shape))
    u = dolfinx.fem.Function(V, dtype=dofs.dtype)
    data_array = u.x.array.reshape(-1, num_components)
    for i in range(num_components):
        local_entities, local_values = dolfinx.io.distribute_entity_data(
            mesh, mesh.topology.dim, topology, dofs[:, i].copy()
        )
        adj = dolfinx.graph.adjacencylist(local_entities)
        order = np.arange(len(local_values), dtype=np.int32)
        mt = dolfinx.mesh.meshtags_from_entities(mesh, mesh.topology.dim, adj, order)
        data_array[mt.indices, i] = local_values[mt.values]
    u.x.scatter_forward()
    return u



"""
Module that can read the VTKHDF format using h5py.
"""

from pathlib import Path
from typing import Any

from mpi4py import MPI

import basix
import dolfinx
import h5py
import numpy as np
import numpy.typing as npt

from io4dolfinx.structures import ArrayData, FunctionData, MeshData, MeshTagsData, ReadMeshData
from io4dolfinx.utils import check_file_exists, compute_local_range

from .. import FileMode, ReadMode
from ..h5py.backend import convert_file_mode, h5pyfile
from ..pyvista.backend import _arbitrary_lagrange_vtk, _cell_degree, _first_order_vtk

read_mode = ReadMode.parallel

_vtk_hdf_version = np.array([2, 1], dtype=np.int32)


def get_default_backend_args(arguments: dict[str, Any] | None) -> dict[str, Any]:
    """Get default backend arguments given a set of input arguments.

    Args:
        arguments: Input backend arguments

    Returns:
        Updated backend arguments
    """
    args = arguments or {"name": "mesh"}
    return args


def find_all_unique_cell_types(comm, cell_types, num_nodes):
    """
    Given a set of cell types and number of nodes per cell, find all unique cell types
    across all ranks.

    Args:
        comm: MPI communicator
        cell_types: Local cell types
        num_nodes: Number of nodes per cell


    Returns:
        A 2D array where each row corresponds to a cell type (vtk int)
        and the number of nodes.
    """
    # Combine cell_types, num_nodes as tuple
    c_hash = np.zeros((2, len(cell_types)), dtype=np.int32)
    c_hash[0] = cell_types
    c_hash[1] = num_nodes
    indexes = np.unique(c_hash.T, axis=0, return_index=True)[1]
    local_unique_cells = c_hash.T[indexes]

    all_cell_types = np.vstack(comm.allgather(local_unique_cells))
    indexes = np.unique(all_cell_types, axis=0, return_index=True)[1]
    all_unique_cell_types = all_cell_types[indexes]
    return all_unique_cell_types


def _decode_bytes_if_needed(value: bytes | str) -> str:
    """Decode bytes to string if necessary (for h5py compatibility)"""
    if isinstance(value, bytes):
        return value.decode("utf-8")
    return value


def _get_vtk_group(h5file, name: str) -> h5py.Group:
    """
    Navigates the VTKHDF group hierarchy to find the specific first
    UnstructuredGrid with a specific name.
    Handles both MultiBlockDataSet and direct UnstructuredGrid types.
    """
    hdf = h5file["VTKHDF"]
    file_type = _decode_bytes_if_needed(hdf.attrs["Type"])

    if file_type == "MultiBlockDataSet":
        ass = hdf["Assembly"]

        def visitor(path):
            mesh_group = path.rsplit("/", 1)
            mesh_name = mesh_group[0] if len(mesh_group) == 1 else mesh_group[1]
            if mesh_name == name:
                obj = ass.get(path)
                # Check attributes carefully
                if "Type" in obj.attrs.keys():
                    attr_type = obj.attrs["Type"]
                    if isinstance(attr_type, bytes):
                        attr_type = attr_type.decode("utf-8")
                    if attr_type == "UnstructuredGrid":
                        return path
            return None

        mesh_node = ass.visit_links(visitor)
        if mesh_node is None:
            raise RuntimeError(f"Could not find unique mesh named '{name}' in Assembly.")

        return ass[mesh_node]

    elif file_type == "UnstructuredGrid":
        return hdf
    else:
        raise RuntimeError(f"Not supported file type {file_type}")


def _get_time_index(hdf: h5py.Group, time: float | str, filename: str | Path) -> int:
    """Finds the index of a specific time stamp."""
    if "Steps" not in hdf.keys():
        raise RuntimeError(f"No timestepping information found in {filename}.")
    stamps = hdf["Steps"]["Values"][:]
    pos = np.flatnonzero(np.isclose(stamps, time))
    if len(pos) == 0:
        raise RuntimeError(f"Could not find mesh at t={time} in {filename}.")
    elif len(pos) > 1:
        raise RuntimeError(f"Multiple time steps for mesh at t={time} in {filename}")
    return pos[0]


def read_mesh_data(
    filename: Path | str,
    comm: MPI.Intracomm,
    time: str | float | None,
    read_from_partition: bool,
    backend_args: dict[str, Any] | None,
) -> ReadMeshData:
    """Read mesh data from file.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        time: Time stamp associated with the mesh to read
        read_from_partition: Whether to read partition information
        backend_args: Arguments to backend

    Returns:
        Internal data structure for the mesh data read from file
    """
    backend_args = get_default_backend_args(backend_args)
    check_file_exists(filename)
    if read_from_partition:
        raise RuntimeError("Cannot read partition data with VTKHDF")
    with h5pyfile(filename, "r", comm=comm) as h5file:
        hdf = _get_vtk_group(h5file, backend_args["name"])
        if time is None:
            num_cells_global = hdf["Types"].size
            local_cell_range = compute_local_range(comm, num_cells_global)
            cell_types_local = hdf["Types"][slice(*local_cell_range)]

            num_points_global = hdf["NumberOfPoints"][0]
            local_point_range = compute_local_range(comm, num_points_global)
            points_local = hdf["Points"][slice(*local_point_range)]

            # Connectivity read
            offsets = hdf["Offsets"]
            local_connectivity_offset = offsets[local_cell_range[0] : local_cell_range[1] + 1]
            topology = hdf["Connectivity"][
                local_connectivity_offset[0] : local_connectivity_offset[-1]
            ]
            offset = local_connectivity_offset - local_connectivity_offset[0]
        else:
            time_index = _get_time_index(hdf, time, filename)

            stamps = hdf["Steps"]["Values"][:]
            # Get number of points
            point_node = hdf["Points"]
            step_start = hdf["Steps"]["PointOffsets"][time_index]

            # NOTE: currently, it doesn't seem like we follow:
            # https://docs.vtk.org/en/latest/vtk_file_formats/vtkhdf_file_format/vtkhdf_specifications.html#temporal-unstructuredgrid-and-polydata
            # As only one num_points is stored irregardless of time steps added.
            if hdf["NumberOfPoints"].shape[0] != len(stamps):
                num_pts = hdf["NumberOfPoints"][0]
            else:
                num_pts = hdf["NumberOfPoints"][time_index]
            lr = compute_local_range(comm, num_pts)
            points_local = point_node[step_start + lr[0] : step_start + lr[1]]

            # Get cell-types in step
            cell_start = hdf["Steps"]["CellOffsets"][time_index]
            if hdf["NumberOfCells"].shape[0] != len(stamps):
                num_cells = hdf["NumberOfCells"][0]
            else:
                num_cells = hdf["NumberOfCells"][time_index]
            local_cell_range = compute_local_range(comm, num_cells)
            cell_types_local = hdf["Types"][
                cell_start + local_cell_range[0] : cell_start + local_cell_range[1]
            ]

            # Get connectivity in step
            connectivity_start = hdf["Steps"]["ConnectivityIdOffsets"][time_index]
            # Connectivity read
            offsets = hdf["Offsets"]
            local_connectivity_offset = offsets[
                connectivity_start + local_cell_range[0] : connectivity_start
                + local_cell_range[1]
                + 1
            ]
            topology = hdf["Connectivity"][
                local_connectivity_offset[0] : local_connectivity_offset[-1]
            ]
            offset = local_connectivity_offset - local_connectivity_offset[0]

    # NOTE: Currently we limit ourselfs to a single celltype, as it makes life easier,
    # other things have to change in `MeshReadData` to support this.
    num_nodes_per_cell = offset[1:] - offset[:-1]
    unique_cells = find_all_unique_cell_types(MPI.COMM_WORLD, cell_types_local, num_nodes_per_cell)
    if unique_cells.shape[0] > 1:
        raise NotImplementedError("io4dolfinx does not support mixed celltype grids")
    topology = topology.reshape(-1, num_nodes_per_cell[0])
    cell_type, number_of_nodes = unique_cells[0]
    gtype = backend_args.get("dtype", points_local.dtype)
    if cell_type in _first_order_vtk.keys():
        ct = _first_order_vtk[cell_type]
        degree = 1
    elif cell_type in _arbitrary_lagrange_vtk.keys():
        ct = _arbitrary_lagrange_vtk[cell_type]
        degree = _cell_degree(ct, num_nodes=number_of_nodes)
    else:
        raise ValueError(f"Unknown VTK cell type {cell_type} in {filename}")
    perm = dolfinx.cpp.io.perm_vtk(dolfinx.mesh.to_type(ct), number_of_nodes)
    topology = topology[:, perm]
    lvar = int(basix.LagrangeVariant.equispaced)
    return ReadMeshData(
        cells=topology, cell_type=ct, x=points_local.astype(gtype), lvar=lvar, degree=degree
    )


def read_point_data(
    filename: Path | str,
    name: str,
    comm: MPI.Intracomm,
    time: float | str | None,
    backend_args: dict[str, Any] | None,
) -> tuple[np.ndarray, int]:
    """Read data from the nodes of a mesh.

    Args:
        filename: Path to file
        name: Name of point data
        comm: Communicator to launch IO on.
        time: The time stamp
        backend_args: The backend arguments

    Returns:
       Data local to process (contiguous, no mpi comm) and local start range
    """
    backend_args = get_default_backend_args(backend_args)
    check_file_exists(filename)
    with h5pyfile(filename, "r", comm=comm) as h5file:
        hdf = _get_vtk_group(h5file, backend_args["name"])
        point_data = hdf["PointData"]
        assert point_data is not None
        if name not in point_data.keys():
            raise ValueError(f"No point data named {name} in {filename}.")
        func_node = point_data[name]

        if time is None:
            data_shape = func_node.shape[0]
            lr = compute_local_range(comm, data_shape)
            data = func_node[slice(*lr)]
            return data, lr[0]
        else:
            time_index = _get_time_index(hdf, time, filename)
            stamps = hdf["Steps"]["Values"][:]
            step_start = hdf["Steps"]["PointDataOffsets"][name][time_index]
            # NOTE: currently, it doesn't seem like we follow:
            # https://docs.vtk.org/en/latest/vtk_file_formats/vtkhdf_file_format/vtkhdf_specifications.html#temporal-unstructuredgrid-and-polydata
            # As only one num_points is stored irregardless of time steps added.
            if hdf["NumberOfPoints"].shape[0] != len(stamps):
                num_pts = hdf["NumberOfPoints"][0]
            else:
                num_pts = hdf["NumberOfPoints"][time_index]
            lr = compute_local_range(comm, num_pts)
            return func_node[step_start + lr[0] : step_start + lr[1]], lr[0]


def read_cell_data(
    filename: Path | str,
    name: str,
    comm: MPI.Intracomm,
    time: str | float | None,
    backend_args: dict[str, Any] | None,
) -> tuple[npt.NDArray[np.int64], np.ndarray]:
    backend_args = get_default_backend_args(backend_args)
    check_file_exists(filename)
    with h5pyfile(filename, "r", comm=comm) as h5file:
        hdf = _get_vtk_group(h5file, backend_args["name"])

        if "CellData" not in hdf.keys():
            raise RuntimeError(f"No cell data found in {filename}.")
        cell_data = hdf["CellData"]
        assert cell_data is not None
        if name not in cell_data.keys():
            raise ValueError(f"No cell data with name {name} in {filename}")
        cell_data_node = cell_data[name]
        assert cell_data_node is not None
        if time is None:
            cell_data_shape = cell_data_node.shape
            num_cells_global = hdf["Types"].size
            assert num_cells_global == cell_data_shape[0]
            local_cell_range = compute_local_range(comm, num_cells_global)
            data = cell_data_node[slice(*local_cell_range)]
        else:
            time_index = _get_time_index(hdf, time, filename)
            stamps = hdf["Steps"]["Values"][:]
            cd_start = hdf["Steps"]["CellDataOffsets"][name][time_index]

            # NOTE: currently, it doesn't seem like we follow:
            # https://docs.vtk.org/en/latest/vtk_file_formats/vtkhdf_file_format/vtkhdf_specifications.html#temporal-unstructuredgrid-and-polydata
            # As only one num_points is stored irregardless of time steps added.
            if hdf["NumberOfCells"].shape[0] != len(stamps):
                number_of_cells = hdf["NumberOfCells"][0]
            else:
                number_of_cells = hdf["NumberOfCells"][time_index]
            lr = compute_local_range(comm, number_of_cells)
            data = cell_data_node[cd_start + lr[0] : cd_start + lr[1]]

    # NOTE: THis could be optimized by hand-coding some communication in
    # `read_cell_data` on the frontend side
    md = read_mesh_data(
        filename, comm, time=time, read_from_partition=False, backend_args=backend_args
    )
    if len(data.shape) == 1:
        data = data.reshape(-1, 1)
    return md.cells, data


def write_attributes(
    filename: Path | str,
    comm: MPI.Intracomm,
    name: str,
    attributes: dict[str, np.ndarray],
    backend_args: dict[str, Any] | None,
):
    """Write attributes to file.

    Args:
        filename: Path to file to write to
        comm: MPI communicator used in storage
        name: Name of the attribute group
        attributes: Dictionary of attributes to write
        backend_args: Arguments to backend
    """
    raise NotImplementedError("The Pyvista backend cannot write attributes.")


def read_attributes(
    filename: Path | str,
    comm: MPI.Intracomm,
    name: str,
    backend_args: dict[str, Any] | None,
) -> dict[str, Any]:
    """Read attributes from file.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the attribute group
        backend_args: Arguments to backend

    Returns:
        Dictionary of attributes read from file
    """
    raise NotImplementedError("The Pyvista backend cannot read attributes.")


def read_timestamps(
    filename: Path | str,
    comm: MPI.Intracomm,
    function_name: str,
    backend_args: dict[str, Any] | None,
) -> npt.NDArray[np.float64 | str]:  # type: ignore[type-var]
    """Read timestamps from file.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        function_name: Name of the function to read timestamps for
        backend_args: Arguments to backend

    Returns:
        Numpy array of timestamps read from file
    """
    backend_args = get_default_backend_args(backend_args)
    check_file_exists(filename)
    # Temporal data storage as described in
    # https://docs.vtk.org/en/latest/vtk_file_formats/vtkhdf_file_format/vtkhdf_specifications.html#temporal-data
    with h5pyfile(filename, "r", comm=comm) as h5file:
        hdf = h5file["VTKHDF"]
        if "Steps" in hdf.keys():
            timestamps = hdf["Steps"]["Values"][:]
            # For either point data or cell data, check if function exists,
            # and check if offsets in time are changing between steps.
            if "CellData" in hdf.keys() and function_name in hdf["CellData"].keys():
                offsets = hdf["Steps"]["CellDataOffsets"][function_name]
                step_offsets = offsets[:]
                step_diff = np.flatnonzero(step_offsets[1:] - step_offsets[:-1])
                return timestamps[step_diff]
            elif "PointData" in hdf.keys() and function_name in hdf["PointData"].keys():
                offsets = hdf["Steps"]["PointDataOffsets"][function_name]
                step_offsets = offsets[:]
                step_diff = np.flatnonzero(step_offsets[1:] - step_offsets[:-1])
                # This only finds when the offset changes, does not capture first step
                return np.hstack([[timestamps[0]], timestamps[step_diff]])
            else:
                raise RuntimeError(f"Function {function_name} is not assoicated with a time-stamp.")
        else:
            raise RuntimeError(f"{filename} does not contain time-stepping information.")


def read_function_names(
    filename: Path | str, comm: MPI.Intracomm, backend_args: dict[str, Any] | None
) -> list[str]:
    """Read all function names from a file.

    Args:
        filename: Path to file
        comm: MPI communicator to launch IO on.
        backend_args: Arguments to backend

    Returns:
        A list of function names.
    """
    backend_args = get_default_backend_args(backend_args)
    check_file_exists(filename)
    with h5pyfile(filename, "r", comm=comm) as h5file:
        hdf = h5file["VTKHDF"]
        function_names = set()
        if "CellData" in hdf.keys():
            for item in hdf["CellData"].keys():
                function_names.add(item)
        if "PointData" in hdf.keys():
            for item in hdf["PointData"].keys():
                function_names.add(item)
    return list(function_names)


def _create_dataset(
    root: h5py.File | h5py.Group,
    name: str,
    shape: tuple[int, ...],
    dtype: npt.DTypeLike,
    chunks: bool,
    maxshape: tuple[int | None, ...],
    mode: str,
    resize: bool = True,
) -> h5py.Dataset:
    if name not in root.keys():
        mode = "w"
    if mode == "w":
        dataset = root.create_dataset(
            name, shape=shape, dtype=dtype, chunks=chunks, maxshape=maxshape
        )
    elif mode == "a":
        dataset = root[name]
        old_shape = dataset.shape
        # Only resize for dimension
        if resize:
            if len(old_shape) == 1:
                new_shape = (old_shape[0] + shape[0],)
            else:
                new_shape = (old_shape[0] + shape[0], *old_shape[1:])
            dataset.resize(new_shape)
    else:
        raise ValueError(f"Unknown file mode '{mode}' when creating dataset {name} in {root}")
    return dataset


def _create_group(root: h5py.File | h5py.Group, name: str, mode: str) -> h5py.Group:
    if name not in root.keys():
        mode = "w"
    if mode == "w":
        # Track order has to be on to make multiblock work:
        # https://docs.vtk.org/en/latest/vtk_file_formats/vtkhdf_file_format/vtkhdf_specifications.html#partitioneddatasetcollection-and-multiblockdataset
        group = root.create_group(name, track_order=True)
    elif mode == "a":
        group = root[name]
    else:
        raise ValueError("Unknown file mode '{h5_mode}'")
    return group


def _compute_append_slice(
    dataset: h5py.Dataset, input_size: int, original_slice: tuple[int, int] | np.ndarray, mode: str
) -> slice:
    append_offset = dataset.shape[0] - input_size if mode == "a" else 0
    return slice(*(np.asarray(original_slice) + append_offset).astype(np.int64))


def write_mesh(
    filename: Path | str,
    comm: MPI.Intracomm,
    mesh: MeshData,
    backend_args: dict[str, Any] | None,
    mode: FileMode,
    time: float,
):
    """
    Write a mesh to file.

    Args:
        comm: MPI communicator used in storage
        mesh: Internal data structure for the mesh data to save to file
        filename: Path to file to write to
        backend_args: Arguments to backend
        mode: File-mode to store the mesh
        time: Time stamp associated with the mesh
    """
    h5_mode = convert_file_mode(mode)
    backend_args = get_default_backend_args(backend_args)
    name = backend_args["name"]

    with h5pyfile(filename, h5_mode, comm=comm) as h5file:
        hdf = _create_group(h5file, "/VTKHDF", h5_mode)
        hdf.attrs.create("Type", "MultiBlockDataSet")
        hdf.attrs["Version"] = _vtk_hdf_version

        mesh_group = _create_group(hdf, name, h5_mode)
        mesh_group.attrs.create("Type", "UnstructuredGrid")
        mesh_group.attrs["Version"] = _vtk_hdf_version

        assembly = _create_group(hdf, "Assembly", h5_mode)

        mesh_assembly = _create_group(assembly, name, h5_mode)
        if name not in mesh_assembly.keys():
            mesh_assembly[name] = h5py.SoftLink(f"/VTKHDF/{name}")

        # Write time dependent points
        number_of_points = _create_dataset(
            mesh_group,
            "NumberOfPoints",
            shape=(1,),
            dtype=np.int64,
            chunks=True,
            maxshape=(None,),
            mode=h5_mode,
        )
        number_of_points[-1] = mesh.num_nodes_global

        # Store nodes
        points = _create_dataset(
            mesh_group,
            "Points",
            shape=(mesh.num_nodes_global, 3),
            dtype=mesh.local_geometry.dtype,
            chunks=True,
            maxshape=(None, 3),
            mode=h5_mode,
        )
        insert_slice = _compute_append_slice(
            points, mesh.num_nodes_global, mesh.local_geometry_pos, h5_mode
        )
        points[insert_slice, : mesh.local_geometry.shape[1]] = mesh.local_geometry

        # NOTE: VTKHDF currently does not support reading time dependent topology in
        # Paraview: https://gitlab.kitware.com/vtk/vtk/-/issues/19257
        # Therefore the following data is not resized
        time_indep_datasets = {}
        for key in ["NumberOfCells", "NumberOfConnectivityIds"]:
            time_indep_datasets[key] = _create_dataset(
                mesh_group,
                key,
                shape=(1,),
                dtype=np.int64,
                chunks=True,
                maxshape=(None,),
                mode=h5_mode,
                resize=False,  # Resize should really be True
            )

        num_dofs_per_cell = mesh.local_topology.shape[1]
        time_indep_datasets["NumberOfCells"][-1] = mesh.num_cells_global
        time_indep_datasets["NumberOfConnectivityIds"][-1] = (
            mesh.num_cells_global * num_dofs_per_cell
        )

        # NOTE: The following offsets are currently overwriting the existing topology data.
        # This is due to the VTKHDF bug. When we switch resize=True this will automatically work
        # for time dependent topologies

        # Store topology offsets (single celltype assumption)
        offsets = _create_dataset(
            mesh_group,
            "Offsets",
            shape=(mesh.num_cells_global + 1,),
            dtype=np.int64,
            chunks=True,
            mode=h5_mode,
            maxshape=(None,),
            resize=False,  # Resize should really be True
        )
        offset_data = np.arange(0, mesh.local_topology.size + 1, mesh.local_topology.shape[1])
        offset_data += num_dofs_per_cell * mesh.local_topology_pos[0]
        insert_slice = _compute_append_slice(
            offsets,
            mesh.num_cells_global + 1,
            (mesh.local_topology_pos[0], mesh.local_topology_pos[1] + 1),
            mode=h5_mode,
        )
        offsets[insert_slice] = offset_data
        del offset_data

        # Permute and store topology data
        dx_ct = dolfinx.mesh.to_type(mesh.cell_type)
        top_perm = np.argsort(dolfinx.cpp.io.perm_vtk(dx_ct, num_dofs_per_cell))
        topology_data = mesh.local_topology[:, top_perm].flatten()
        topology = _create_dataset(
            mesh_group,
            "Connectivity",
            shape=(mesh.num_cells_global * num_dofs_per_cell,),
            dtype=np.int64,
            chunks=True,
            maxshape=(None,),
            mode=h5_mode,
            resize=False,  # Resize should really be True, see issue below
        )  # VTKHDFReader issue: https://gitlab.kitware.com/vtk/vtk/-/issues/19257
        insert_slice = _compute_append_slice(
            topology,
            mesh.num_cells_global * num_dofs_per_cell,
            np.array(mesh.local_topology_pos) * num_dofs_per_cell,
            mode=h5_mode,
        )
        topology[insert_slice] = topology_data
        del topology_data

        # Store celltypes
        cell_types = np.full(
            mesh.local_topology.shape[0],
            dolfinx.cpp.io.get_vtk_cell_type(dx_ct, dolfinx.mesh.cell_dim(dx_ct)),
        )
        types = _create_dataset(
            mesh_group,
            "Types",
            shape=(mesh.num_cells_global,),
            dtype=np.uint8,
            maxshape=(None,),
            chunks=True,
            mode=h5_mode,
            resize=False,  # Resize should really be True, see issue below
        )  # VTKHDFReader issue: https://gitlab.kitware.com/vtk/vtk/-/issues/19257
        insert_slice = _compute_append_slice(
            types, mesh.num_cells_global, mesh.local_topology_pos, h5_mode
        )
        types[insert_slice] = cell_types
        del cell_types

        steps = _create_group(mesh_group, "Steps", mode=h5_mode)
        # First fetch time-steps to see if we have stored this timestep already
        values = _create_dataset(
            steps,
            "Values",
            shape=(1,),
            dtype=np.float64,
            chunks=True,
            maxshape=(None,),
            mode=h5_mode,
            resize=h5_mode == "a",
        )

        if h5_mode == "w":
            values[0] = time
        else:
            existing_steps = values[:-1]
            if len(np.flatnonzero(np.isclose(existing_steps, time))) > 0:
                raise RuntimeError(f"Mesh already exists at time {time} in {filename}.")
            values[-1] = time
        steps.attrs.create("NSteps", np.int64(len(values)), dtype=np.int64)

        # Write offset data for current time-step
        all_parts = {}
        for key in [
            "NumberOfParts",
            "PartOffsets",
            "PointOffsets",
            "CellOffsets",
            "ConnectivityIdOffsets",
        ]:
            all_parts[key] = _create_dataset(
                steps,
                key,
                shape=(1,),
                dtype=np.int64,
                chunks=True,
                maxshape=(None,),
                mode=h5_mode,
            )

        all_parts["NumberOfParts"][-1] = 1
        all_parts["PartOffsets"][-1] = 0
        all_parts["PointOffsets"][-1] = points.shape[0] - mesh.num_nodes_global
        all_parts["CellOffsets"][-1] = types.shape[0] - mesh.num_cells_global
        all_parts["ConnectivityIdOffsets"][-1] = (
            topology.shape[0] - mesh.num_cells_global * num_dofs_per_cell
        )

        # Update cell-data and point-data offsets by copying over data from previous step
        for key in ["CellDataOffsets", "PointDataOffsets"]:
            group = _create_group(steps, key, mode=h5_mode)
            for cd in group.keys():
                group[cd].resize(steps.attrs["NSteps"], axis=0)
                group[cd][-1] = group[cd][-2]

        # Update Steps in all other parts of the mesh as well
        for key in mesh_assembly.keys():
            if key == name:
                continue
            # Copy time-dependent geometry info (NumberOfPoints) from mesh to tag
            sub_group = mesh_assembly[key]
            sub_step = sub_group["Steps"]
            sub_step.attrs["NSteps"] = steps.attrs["NSteps"]

            # Copy time dependent and partition info from mesh to tag
            step_copy_keys = ["Values", "PartOffsets", "NumberOfParts"]
            for key in step_copy_keys:
                if key in sub_step.keys():
                    sub_step[key].resize(steps[key].shape)
                    sub_step[key][:] = steps[key][:]
                else:
                    raise RuntimeError(f"{sub_step.name} should have {key}/")

            # Append value from previous step for meshtags as they are time-independent
            append_keys = ["CellOffsets", "ConnectivityIdOffsets"]
            for key in append_keys:
                if key in sub_step.keys():
                    sub_step[key].resize(sub_step[key].size + 1, axis=0)
                    sub_step[key][-1] = sub_step[key][-2]
                else:
                    raise RuntimeError(f"{sub_step.name} should have {key}/")
            # Append value from previous step for meshtags celldata
            for key in ["CellDataOffsets", "PointDataOffsets"]:
                group = _create_group(sub_step, key, mode=h5_mode)
                for cd in group.keys():
                    group[cd].resize(sub_step.attrs["NSteps"], axis=0)
                    group[cd][-1] = group[cd][-2]


def write_meshtags(
    filename: str | Path,
    comm: MPI.Intracomm,
    data: MeshTagsData,
    backend_args: dict[str, Any] | None,
):
    """Write mesh tags to file.

    Args:
        filename: Path to file to write to
        comm: MPI communicator used in storage
        data: Internal data structure for the mesh tags to save to file
        backend_args: Arguments to backend
    """
    backend_args = get_default_backend_args(backend_args)
    name = backend_args["name"]
    h5_mode = "a"
    with h5pyfile(filename, h5_mode, comm=comm) as h5file:
        hdf = h5file["VTKHDF"]
        file_type = _decode_bytes_if_needed(hdf.attrs["Type"])
        # Check for type of VTKHDF file
        if file_type != "MultiBlockDataSet":
            raise ValueError(f"Cannot write meshtags to {filename} with VTK type {file_type}")

        # We place the mesh-tags in the same subgroup of assembly as the mesh
        parent_mesh_group = _get_vtk_group(h5file, name)
        block = parent_mesh_group.parent

        tag_path = f"/VTKHDF/{name}_{data.name}"
        if data.name not in block.keys():
            block[data.name] = h5py.SoftLink(tag_path)

        mesh_group = _create_group(hdf, tag_path, mode=h5_mode)
        mesh_group.attrs.create("Type", "UnstructuredGrid")
        mesh_group.attrs["Version"] = _vtk_hdf_version

        cell_data = _create_group(mesh_group, "CellData", mode=h5_mode)
        assert data.num_entities_global is not None
        dataset = _create_dataset(
            cell_data,
            data.name,
            shape=(data.num_entities_global,),
            dtype=data.values.dtype,
            chunks=True,
            maxshape=(None,),
            mode=h5_mode,
        )
        assert data.local_start is not None
        insert_slice = _compute_append_slice(
            dataset,
            data.num_entities_global,
            np.array([data.local_start, data.local_start + data.indices.shape[0]]),
            mode=h5_mode,
        )
        dataset[insert_slice] = data.values

        # NOTE: The following is more or less a copy from write_mesh,
        # except that we pull out the point storage and use a softlink
        num_cells = _create_dataset(
            mesh_group,
            "NumberOfCells",
            shape=(1,),
            dtype=np.int64,
            chunks=True,
            maxshape=(None,),
            mode=h5_mode,
            resize=False,  # Resize should really be True, see issue below
        )  # VTKHDFReader issue: https://gitlab.kitware.com/vtk/vtk/-/issues/19257
        num_cells[-1] = data.num_entities_global

        # Hardlink data should also follow hardlink for numbering
        for key in ["Points", "NumberOfPoints"]:
            if key not in mesh_group.keys():
                mesh_group[key] = parent_mesh_group[key]

        # Single celltype assumption
        num_dofs_per_cell = data.num_dofs_per_entity
        number_of_connectivities = _create_dataset(
            mesh_group,
            "NumberOfConnectivityIds",
            shape=(1,),
            dtype=np.int64,
            chunks=True,
            maxshape=(None,),
            mode=h5_mode,
            resize=False,  # Resize should really be True, see issue below
        )  # VTKHDFReader issue: https://gitlab.kitware.com/vtk/vtk/-/issues/19257
        assert data.num_entities_global is not None and num_dofs_per_cell is not None
        number_of_connectivities[-1] = data.num_entities_global * num_dofs_per_cell

        # Store topology offsets (single celltype assumption)
        offsets = _create_dataset(
            mesh_group,
            "Offsets",
            shape=(data.num_entities_global + 1,),
            dtype=np.int64,
            chunks=True,
            mode=h5_mode,
            maxshape=(None,),
            resize=False,  # Resize should really be True, see issue below
        )  # VTKHDFReader issue: https://gitlab.kitware.com/vtk/vtk/-/issues/19257
        offset_data = np.arange(0, data.indices.size + 1, data.indices.shape[1])
        assert data.local_start is not None
        offset_data += num_dofs_per_cell * data.local_start
        insert_slice = _compute_append_slice(
            offsets,
            data.num_entities_global + 1,
            (data.local_start, data.local_start + data.indices.shape[0] + 1),
            mode=h5_mode,
        )
        offsets[insert_slice] = offset_data
        del offset_data

        # Permute and store topology data
        dx_ct = dolfinx.mesh.to_type(data.cell_type)
        top_perm = np.argsort(dolfinx.cpp.io.perm_vtk(dx_ct, num_dofs_per_cell))
        topology_data = data.indices[:, top_perm].flatten()
        topology = _create_dataset(
            mesh_group,
            "Connectivity",
            shape=(data.num_entities_global * num_dofs_per_cell,),
            dtype=np.int64,
            chunks=True,
            maxshape=(None,),
            mode=h5_mode,
            resize=False,  # Resize should really be True, see issue below
        )  # VTKHDFReader issue: https://gitlab.kitware.com/vtk/vtk/-/issues/19257
        insert_slice = _compute_append_slice(
            topology,
            data.num_entities_global * num_dofs_per_cell,
            np.array([data.local_start, data.local_start + data.indices.shape[0]])
            * num_dofs_per_cell,
            mode=h5_mode,
        )
        topology[insert_slice] = topology_data
        del topology_data

        # Store celltypes
        cell_types = np.full(
            data.indices.shape[0],
            dolfinx.cpp.io.get_vtk_cell_type(dx_ct, dolfinx.mesh.cell_dim(dx_ct)),
        )
        types = _create_dataset(
            mesh_group,
            "Types",
            shape=(data.num_entities_global,),
            dtype=np.uint8,
            maxshape=(None,),
            chunks=True,
            mode=h5_mode,
            resize=False,  # Resize should really be True, see issue below
        )  # VTKHDFReader issue: https://gitlab.kitware.com/vtk/vtk/-/issues/19257
        insert_slice = _compute_append_slice(
            types,
            data.num_entities_global,
            (data.local_start, data.local_start + data.indices.shape[0]),
            h5_mode,
        )
        types[insert_slice] = cell_types
        del cell_types

        steps = _create_group(mesh_group, "Steps", mode=h5_mode)

        # Copy n-step counter
        steps.attrs.create("NSteps", parent_mesh_group["Steps"].attrs["NSteps"])

        # Hardlink data that we know is the same across meshes
        hardlink_keys = ["NumberOfParts", "PartOffsets", "Values", "PointOffsets"]
        for key in hardlink_keys:
            steps[key] = parent_mesh_group["Steps"][key]

        # Write offset data for current time-step
        all_parts = {}
        for key in ["CellOffsets", "ConnectivityIdOffsets"]:
            all_parts[key] = _create_dataset(
                steps,
                key,
                shape=(1,),
                dtype=np.int64,
                chunks=True,
                maxshape=(None,),
                mode=h5_mode,
            )

        all_parts["CellOffsets"][-1] = types.shape[0] - data.num_entities_global
        all_parts["ConnectivityIdOffsets"][-1] = offsets.shape[0] - (data.num_entities_global + 1)
        # CellData requires an offset
        cd_off = _create_group(steps, "CellDataOffsets", mode=h5_mode)
        cd_data = _create_dataset(
            cd_off,
            data.name,
            shape=parent_mesh_group["Steps"]["CellOffsets"].shape,
            dtype=np.int64,
            chunks=True,
            maxshape=(None,),
            mode=h5_mode,
        )
        cd_data[-1] = types.shape[0] - data.num_entities_global


def read_meshtags_data(
    filename: str | Path,
    comm: MPI.Intracomm,
    name: str,
    backend_args: dict[str, Any] | None,
) -> MeshTagsData:
    """Read mesh tags from file.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the mesh tags to read
        backend_args: Arguments to backend

    Returns:
        Internal data structure for the mesh tags read from file
    """
    backend_args = get_default_backend_args(backend_args)
    backend_args.update({"name": name})
    # Reuse reading cell-data
    indices, values = read_cell_data(filename, name, comm, None, backend_args=backend_args)
    # Read cell-type of grid to get topological dimension
    with h5pyfile(filename, "r", comm=comm) as h5file:
        hdf = _get_vtk_group(h5file, backend_args["name"])
        num_cells_global = hdf["Types"].size
        local_cell_range = compute_local_range(comm, num_cells_global)
        cell_types_local = hdf["Types"][slice(*local_cell_range)]
    unique_cells = find_all_unique_cell_types(comm, cell_types_local, indices.shape[1])
    if unique_cells.shape[0] > 1:
        raise NotImplementedError("io4dolfinx does not support mixed celltype grids")
    vtk_cell_type = unique_cells[0][0]
    if vtk_cell_type in _first_order_vtk.keys():
        ct = _first_order_vtk[vtk_cell_type]
    elif vtk_cell_type in _arbitrary_lagrange_vtk.keys():
        ct = _arbitrary_lagrange_vtk[vtk_cell_type]
    dim = dolfinx.mesh.cell_dim(dolfinx.mesh.to_type(ct))
    return MeshTagsData(name=name, values=values.flatten(), indices=indices, dim=dim)


def read_dofmap(
    filename: str | Path,
    comm: MPI.Intracomm,
    name: str,
    backend_args: dict[str, Any] | None,
) -> dolfinx.graph.AdjacencyList:
    """Read the dofmap of a function with a given name.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the function to read the dofmap for
        backend_args: Arguments to backend

    Returns:
        Dofmap as an AdjacencyList
    """
    raise NotImplementedError("The Pyvista backend cannot make checkpoints.")


def read_dofs(
    filename: str | Path,
    comm: MPI.Intracomm,
    name: str,
    time: float,
    backend_args: dict[str, Any] | None,
) -> tuple[npt.NDArray[np.float32 | np.float64 | np.complex64 | np.complex128], int]:
    """Read the dofs (values) of a function with a given name from a given timestep.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the function to read the dofs for
        time: Time stamp associated with the function to read
        backend_args: Arguments to backend

    Returns:
        Contiguous sequence of degrees of freedom (with respect to input data)
        and the global starting point on the process.
        Process 0 has [0, M), process 1 [M, N), process 2 [N, O) etc.
    """
    raise NotImplementedError("The Pyvista backend cannot make checkpoints.")


def read_cell_perms(
    comm: MPI.Intracomm, filename: Path | str, backend_args: dict[str, Any] | None
) -> npt.NDArray[np.uint32]:
    """
    Read cell permutation from file with given communicator,
    Split in continuous chunks based on number of cells in the input data.

    Args:
        comm: MPI communicator used in storage
        filename: Path to file to read from
        backend_args: Arguments to backend

    Returns:
        Contiguous sequence of permutations (with respect to input data)
        Process 0 has [0, M), process 1 [M, N), process 2 [N, O) etc.
    """
    raise NotImplementedError("The VTKHDF backend cannot make checkpoints.")


def write_function(
    filename: Path,
    comm: MPI.Intracomm,
    u: FunctionData,
    time: float,
    mode: FileMode,
    backend_args: dict[str, Any] | None,
):
    """
    Write a function to file.

    Args:
        comm: MPI communicator used in storage
        u: Internal data structure for the function data to save to file
        filename: Path to file to write to
        time: Time stamp associated with function
        mode: File-mode to store the function
        backend_args: Arguments to backend
    """
    raise NotImplementedError("The VTKHDF backend cannot make checkpoints.")


def read_legacy_mesh(
    filename: Path | str, comm: MPI.Intracomm, group: str
) -> tuple[npt.NDArray[np.int64], npt.NDArray[np.floating], str | None]:
    """Read in the mesh topology, geometry and (optionally) cell type from a
    legacy DOLFIN HDF5-file.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        group: Group in HDF5 file where mesh is stored

    Returns:
        Tuple containing:
            - Topology as a (num_cells, num_vertices_per_cell) array of global vertex indices
            - Geometry as a (num_vertices, geometric_dimension) array of vertex coordinates
            - Cell type as a string (e.g. "tetrahedron") or None if not found
    """
    raise NotImplementedError("The VTKHDF backend cannot read legacy DOLFIN meshes.")


def snapshot_checkpoint(
    filename: Path | str,
    mode: FileMode,
    u: dolfinx.fem.Function,
    backend_args: dict[str, Any] | None,
):
    """Create a snapshot checkpoint of a dolfinx function.

    Args:
        filename: Path to file to read from
        mode: File-mode to store the function
        u: dolfinx function to create a snapshot checkpoint for
        backend_args: Arguments to backend
    """
    raise NotImplementedError("The VTKHDF backend cannot make checkpoints.")


def read_hdf5_array(
    comm: MPI.Intracomm,
    filename: Path | str,
    group: str,
    backend_args: dict[str, Any] | None,
) -> tuple[np.ndarray, int]:
    """Read an array from an HDF5 file.

    Args:
        comm: MPI communicator used in storage
        filename: Path to file to read from
        group: Group in HDF5 file where array is stored
        backend_args: Arguments to backend

    Returns:
        Tuple containing:
            - Numpy array read from file
            - Global starting point on the process.
                Process 0 has [0, M), process 1 [M, N), process 2 [N, O) etc.
    """
    raise NotImplementedError("The VTKHDF backend cannot read HDF5 arrays")


def write_data(
    filename: Path | str,
    array_data: ArrayData,
    comm: MPI.Intracomm,
    time: str | float | None,
    mode: FileMode,
    backend_args: dict[str, Any] | None,
):
    """Write function to file by interpolating into geometry nodes.


    Args:
        filename: Path to file
        array_data: Data to write to file
        time: Time stamp
        mode: Append or write
        backend_args: The backend arguments
    """
    h5_mode = convert_file_mode(mode)
    assert h5_mode == "a"
    backend_args = get_default_backend_args(backend_args)
    mesh_name = backend_args["name"]
    extension = array_data.type
    with h5pyfile(filename, h5_mode, comm=comm) as h5file:
        hdf = h5file["VTKHDF"]
        file_type = _decode_bytes_if_needed(hdf.attrs["Type"])
        # Check for type of VTKHDF file
        if file_type != "MultiBlockDataSet":
            raise ValueError(f"Cannot write meshtags to {filename} with VTK type {file_type}")

        # Find mesh block to add data to
        block = _get_vtk_group(h5file, mesh_name)

        data_group = _create_group(block, f"{extension}Data", mode=h5_mode)
        dataset = _create_dataset(
            data_group,
            array_data.name,
            shape=array_data.global_shape,
            dtype=array_data.values.dtype,
            chunks=True,
            maxshape=(None, array_data.values.shape[1]),
            mode=h5_mode,
        )
        insert_slice = _compute_append_slice(
            dataset,
            array_data.global_shape[0],
            np.array(
                [
                    array_data.local_range[0],
                    array_data.local_range[0] + array_data.values.shape[0],
                ]
            ),
            mode=h5_mode,
        )
        dataset[insert_slice] = array_data.values

        steps = _create_group(block, "Steps", mode=h5_mode)
        pdo = _create_group(steps, f"{extension}DataOffsets", mode=h5_mode)

        # Check if time step is already in time-stepping of mesh
        timestamps = steps["Values"][:]
        assert isinstance(timestamps, np.ndarray)
        assert isinstance(time, float)
        time_exists = np.flatnonzero(np.isclose(timestamps, time))
        if len(time_exists) > 1:
            raise ValueError("Time-step has been written multiple times")
        pdo_u = _create_dataset(
            pdo,
            array_data.name,
            shape=(1,),
            dtype=np.int64,
            chunks=True,
            maxshape=(None,),
            mode=h5_mode,
            resize=False,
        )
        if len(time_exists) == 1:
            idx = time_exists[0]
        elif len(time_exists) == 0:
            # No mesh written at step, update mesh offsets
            # Geometry has to be time-dependent
            num_points = block["NumberOfPoints"]
            num_points.resize(len(timestamps) + 1, axis=0)
            num_points[-1] = num_points[-2]
            # Even if topology cannot be time dep at the moment, number of cells
            # must be updated
            num_cells = block["NumberOfCells"]
            num_cells.resize(len(timestamps) + 1, axis=0)
            num_cells[-1] = num_cells[-2]

            steps.attrs.create("NSteps", block["Steps"].attrs["NSteps"] + 1)
            step_vals = _create_dataset(
                steps,
                "Values",
                shape=(1,),
                dtype=np.float64,
                chunks=True,
                maxshape=(None,),
                mode=h5_mode,
                resize=True,
            )
            step_vals[-1] = time
            idx = step_vals.size - 1
            for key in [
                "PartOffsets",
                "NumberOfParts",
                "PointOffsets",
                "CellOffsets",
                "ConnectivityIdOffsets",
            ]:
                comp = steps[key]
                comp.resize(steps.attrs["NSteps"], axis=0)
                if comp.size > 1:
                    comp[-1] = comp[-2]
        else:
            raise ValueError(f"Time step found multiple times in {filename}")

        # Update offsets for current variable
        pdo_u.resize(steps.attrs["NSteps"], axis=0)
        pdo_u[idx] = dataset.shape[0] - array_data.global_shape[0]

        # Point and cell data of all other variables has to be updated to be synced
        for key in [
            "CellDataOffsets",
            "PointDataOffsets",
        ]:
            comp = steps[key]
            num_steps = steps.attrs["NSteps"]
            for dname, dset in comp.items():
                # Only resize other data arrays
                if dname != array_data.name:
                    # Only resize if it hasn't already been resized
                    if dset.size != num_steps:
                        dset.resize(num_steps, axis=0)
                        # Only update data if we are further than first time step
                        if dset.size > 1:
                            dset[-1] = dset[-2]



# # Snapshot checkpoint (non-persistent)
# The checkpoint method described in [Writing function checkpoints](./writing_functions_checkpoint)
# are *N-to-M*, meaning that you can write them out on N-processes and read them in on M processes.
#
# As discussed in that chapter, these checkpoints need to be associated with a mesh.
# This is because the function is defined on a specific function space, which in turn is
# defined on a specific mesh.
#
# However, there are certain scenarios where you simply want to store a checkpoint associated
# with the current mesh, that should only be possible to use during this simulation.
# An example use-case is when running an iterative solver, and wanting a fall-back mechanism that
# does not require extra RAM.

# In this example, we will demonstrate how to write a
# {py:func}`snapshot checkpoint<io4dolfinx.snapshot_checkpoint>` to disk.

# First we define a {py:class}`function<dolfinx.fem.Function>` `f` that we want to represent in
# the {py:class}`function space<dolfinx.fem.FunctionSpace>`.

# +
import logging
from pathlib import Path

import ipyparallel as ipp


def f(x):
    import numpy as np

    return np.sin(x[0]) + 0.1 * x[1]


# -

# Next, we create a mesh and an appropriate function space and read and write from file.
# Note that for both these operations, we use {py:func}`io4dolfinx.snapshot_checkpoint`,
# with different read and write modes.


def read_write_snapshot(filename: Path):
    from mpi4py import MPI

    import dolfinx
    import numpy as np

    import io4dolfinx

    mesh = dolfinx.mesh.create_unit_cube(MPI.COMM_WORLD, 3, 7, 4)
    V = dolfinx.fem.functionspace(mesh, ("Lagrange", 5))
    u = dolfinx.fem.Function(V)
    u.interpolate(f)
    u.name = "Current_solution"
    # Next, we store the solution to file
    io4dolfinx.snapshot_checkpoint(u, filename, io4dolfinx.FileMode.write)

    # Next, we create a new function and load the solution into it
    u_new = dolfinx.fem.Function(V)
    u_new.name = "Read_solution"
    io4dolfinx.snapshot_checkpoint(u_new, filename, io4dolfinx.FileMode.read)

    # Next, we verify that the solution is correct
    np.testing.assert_allclose(u_new.x.array, u.x.array, atol=np.finfo(float).eps)

    print(f"{MPI.COMM_WORLD.rank + 1}/{MPI.COMM_WORLD.size}: Successfully wrote and read snapshot")


# +

mesh_file = Path("snapshot.bp")

with ipp.Cluster(engines="mpi", n=3, log_level=logging.ERROR) as cluster:
    cluster[:].push({"f": f})
    query = cluster[:].apply_async(
        read_write_snapshot,
        mesh_file,
    )
    query.wait()
    assert query.successful(), query.error
    print("".join(query.stdout))
# -



import warnings
from pathlib import Path
from typing import Any

from mpi4py import MPI

import adios2
import dolfinx
import numpy as np
import numpy.typing as npt

from ...structures import ArrayData, FunctionData, MeshData, MeshTagsData, ReadMeshData
from ...utils import check_file_exists, compute_local_range
from .. import FileMode, ReadMode
from .helpers import (
    ADIOSFile,
    adios_to_numpy_dtype,
    check_variable_exists,
    read_adjacency_list,
    read_array,
    resolve_adios_scope,
)

adios2 = resolve_adios_scope(adios2)

read_mode = ReadMode.parallel


def get_default_backend_args(arguments: dict[str, Any] | None) -> dict[str, Any]:
    """Get default arguements (sets engine to BP4)."""
    args = arguments or {}
    if "engine" not in args.keys():
        args["engine"] = "BP4"
    if "legacy" not in args.keys():
        args["legacy"] = False  # Only used for legacy HDF5 meshtags
    return args


def convert_file_mode(mode: FileMode) -> adios2.Mode:  # type: ignore[override]
    match mode:
        case FileMode.append:
            return adios2.Mode.Append
        case FileMode.write:
            return adios2.Mode.Write
        case FileMode.read:
            return adios2.Mode.Read
        case _:
            raise NotImplementedError(f"FileMode {mode} not implemented.")


def write_attributes(
    filename: Path | str,
    comm: MPI.Intracomm,
    name: str,
    attributes: dict[str, np.ndarray],
    backend_args: dict[str, Any] | None = None,
):
    """Write attributes to file using ADIOS2.

    Args:
        filename: Path to file to write to
        comm: MPI communicator used in storage
        name: Name of the attributes
        attributes: Dictionary of attributes to write to file
        engine: ADIOS2 engine to use
    """

    adios = adios2.ADIOS(comm)
    backend_args = get_default_backend_args(backend_args)

    with ADIOSFile(
        adios=adios,
        filename=filename,
        mode=adios2.Mode.Append,
        io_name="AttributeWriter",
        engine=backend_args["engine"],
    ) as adios_file:
        adios_file.file.BeginStep()

        for k, v in attributes.items():
            adios_file.io.DefineAttribute(f"{name}_{k}", v)

        adios_file.file.PerformPuts()
        adios_file.file.EndStep()


def read_attributes(
    filename: Path | str,
    comm: MPI.Intracomm,
    name: str,
    backend_args: dict[str, Any] | None = None,
) -> dict[str, np.ndarray]:
    """Read attributes from file using ADIOS2.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the attributes
        engine: ADIOS2 engine to use
    Returns:
        The attributes
    """
    check_file_exists(filename)
    adios = adios2.ADIOS(comm)
    backend_args = get_default_backend_args(backend_args)
    with ADIOSFile(
        adios=adios,
        filename=filename,
        mode=adios2.Mode.Read,
        engine=backend_args["engine"],
        io_name="AttributesReader",
    ) as adios_file:
        adios_file.file.BeginStep()
        attributes = {}
        for k in adios_file.io.AvailableAttributes().keys():
            if k.startswith(f"{name}_"):
                a = adios_file.io.InquireAttribute(k)
                attributes[k[len(name) + 1 :]] = a.Data()
        adios_file.file.EndStep()
    return attributes


def read_timestamps(
    filename: Path | str,
    comm: MPI.Intracomm,
    function_name: str,
    backend_args: dict[str, Any] | None = None,
) -> npt.NDArray[np.float64 | str]:  # type: ignore[type-var]
    """Read time-stamps from a checkpoint file.

    Args:
        comm: MPI communicator
        filename: Path to file
        function_name: Name of the function to read time-stamps for
        backend_args: Arguments for backend, for instance file type.
        backend: What backend to use for writing.
    Returns:
        The time-stamps
    """
    check_file_exists(filename)

    adios = adios2.ADIOS(comm)
    backend_args = get_default_backend_args(backend_args)
    with ADIOSFile(
        adios=adios,
        filename=filename,
        mode=adios2.Mode.Read,
        engine=backend_args["engine"],
        io_name="TimestepReader",
    ) as adios_file:
        time_name = f"{function_name}_time"
        time_stamps = []
        for _ in range(adios_file.file.Steps()):
            adios_file.file.BeginStep()
            if time_name in adios_file.io.AvailableVariables().keys():
                arr = adios_file.io.InquireVariable(time_name)
                time_shape = arr.Shape()
                arr.SetSelection([[0], [time_shape[0]]])
                times = np.empty(
                    time_shape[0],
                    dtype=adios_to_numpy_dtype[arr.Type()],
                )
                adios_file.file.Get(arr, times, adios2.Mode.Sync)
                time_stamps.append(times[0])
            adios_file.file.EndStep()

    return np.array(time_stamps)


def write_mesh(
    filename: Path | str,
    comm: MPI.Intracomm,
    mesh: MeshData,
    backend_args: dict[str, Any] | None = None,
    mode: FileMode = FileMode.write,
    time: float = 0.0,
):
    """Write a mesh to file using ADIOS2.

    Args:
        comm: MPI communicator used in storage
        mesh: Internal data structure for the mesh data to save to file
        filename: Path to file to write to
        backend_args: File mode and potentially the io-name.
        mode: Mode to use (write or append)
        time: Time stamp
    """
    backend_args = get_default_backend_args(backend_args)
    if "io_name" not in backend_args.keys():
        backend_args["io_name"] = "MeshWriter"

    mode = convert_file_mode(mode)
    gdim = mesh.local_geometry.shape[1]
    adios = adios2.ADIOS(comm)
    with ADIOSFile(
        adios=adios,
        filename=filename,
        mode=mode,
        comm=comm,
        engine=backend_args["engine"],
        io_name=backend_args["io_name"],
    ) as adios_file:
        adios_file.file.BeginStep()
        # Write geometry
        pointvar = adios_file.io.DefineVariable(
            "Points",
            mesh.local_geometry,
            shape=[mesh.num_nodes_global, gdim],
            start=[mesh.local_geometry_pos[0], 0],
            count=[mesh.local_geometry_pos[1] - mesh.local_geometry_pos[0], gdim],
        )
        adios_file.file.Put(pointvar, mesh.local_geometry, adios2.Mode.Sync)

        if mode == adios2.Mode.Write:
            adios_file.io.DefineAttribute("CellType", mesh.cell_type)
            adios_file.io.DefineAttribute("Degree", np.array([mesh.degree], dtype=np.int32))
            adios_file.io.DefineAttribute(
                "LagrangeVariant", np.array([mesh.lagrange_variant], dtype=np.int32)
            )
            # Write topology (on;y on first write as topology is constant)
            num_dofs_per_cell = mesh.local_topology.shape[1]
            dvar = adios_file.io.DefineVariable(
                "Topology",
                mesh.local_topology,
                shape=[mesh.num_cells_global, num_dofs_per_cell],
                start=[mesh.local_topology_pos[0], 0],
                count=[
                    mesh.local_topology_pos[1] - mesh.local_topology_pos[0],
                    num_dofs_per_cell,
                ],
            )
            adios_file.file.Put(dvar, mesh.local_topology)

            # Add partitioning data
            if mesh.store_partition:
                assert mesh.partition_range is not None
                par_data = adios_file.io.DefineVariable(
                    "PartitioningData",
                    mesh.ownership_array,
                    shape=[mesh.partition_global],
                    start=[mesh.partition_range[0]],
                    count=[
                        mesh.partition_range[1] - mesh.partition_range[0],
                    ],
                )
                adios_file.file.Put(par_data, mesh.ownership_array)
                assert mesh.ownership_offset is not None
                par_offset = adios_file.io.DefineVariable(
                    "PartitioningOffset",
                    mesh.ownership_offset,
                    shape=[mesh.num_cells_global + 1],
                    start=[mesh.local_topology_pos[0]],
                    count=[mesh.local_topology_pos[1] - mesh.local_topology_pos[0] + 1],
                )
                adios_file.file.Put(par_offset, mesh.ownership_offset)
                assert mesh.partition_processes is not None
                adios_file.io.DefineAttribute(
                    "PartitionProcesses", np.array([mesh.partition_processes], dtype=np.int32)
                )
        if mode == adios2.Mode.Append and mesh.store_partition:
            warnings.warn("Partitioning data is not written in append mode")

        # Add time step to file
        t_arr = np.array([time], dtype=np.float64)
        time_var = adios_file.io.DefineVariable(
            "MeshTime",
            t_arr,
            shape=[1],
            start=[0],
            count=[1 if comm.rank == 0 else 0],
        )
        adios_file.file.Put(time_var, t_arr)

        adios_file.file.PerformPuts()
        adios_file.file.EndStep()


def read_mesh_data(
    filename: Path | str,
    comm: MPI.Intracomm,
    time: str | float | None = 0.0,
    read_from_partition: bool = False,
    backend_args: dict[str, Any] | None = None,
) -> ReadMeshData:
    """Read an ADIOS2 mesh data for use with DOLFINx.

    Args:
        filename: Path to input file
        comm: The MPI communciator to distribute the mesh over
        engine: ADIOS engine to use for reading (BP4, BP5 or HDF5)
        time: Time stamp associated with mesh
        legacy: If checkpoint was made prior to time-dependent mesh-writer set to True
        read_from_partition: Read mesh with partition from file
    Returns:
        The mesh topology, geometry, UFL domain and partition function
    """

    adios = adios2.ADIOS(comm)
    backend_args = get_default_backend_args(backend_args)
    legacy = backend_args.get("legacy", False)
    io_name = backend_args.get("io_name", "MeshReader")
    engine = backend_args["engine"]
    with ADIOSFile(
        adios=adios,
        filename=filename,
        mode=adios2.Mode.Read,
        engine=engine,
        io_name=io_name,
    ) as adios_file:
        # Get time independent mesh variables (mesh topology and cell type info) first
        adios_file.file.BeginStep()
        # Get mesh topology (distributed)
        if "Topology" not in adios_file.io.AvailableVariables().keys():
            raise KeyError(f"Mesh topology not found at Topology in {filename}")
        topology = adios_file.io.InquireVariable("Topology")
        shape = topology.Shape()
        local_range = compute_local_range(comm, shape[0])
        topology.SetSelection([[local_range[0], 0], [local_range[1] - local_range[0], shape[1]]])
        mesh_topology = np.empty((local_range[1] - local_range[0], shape[1]), dtype=np.int64)
        adios_file.file.Get(topology, mesh_topology, adios2.Mode.Deferred)

        # Check validity of partitioning information
        if read_from_partition:
            if "PartitionProcesses" not in adios_file.io.AvailableAttributes().keys():
                raise KeyError(f"Partitioning information not found in {filename}")
            par_num_procs = adios_file.io.InquireAttribute("PartitionProcesses")
            num_procs = par_num_procs.Data()[0]
            if num_procs != comm.size:
                raise ValueError(f"Number of processes in file ({num_procs})!=({comm.size=})")

        # Get mesh cell type
        if "CellType" not in adios_file.io.AvailableAttributes().keys():
            raise KeyError(f"Mesh cell type not found at CellType in {filename}")
        celltype = adios_file.io.InquireAttribute("CellType")
        cell_type = celltype.DataString()[0]

        # Get basix info
        if "LagrangeVariant" not in adios_file.io.AvailableAttributes().keys():
            raise KeyError(f"Mesh LagrangeVariant not found in {filename}")
        lvar = adios_file.io.InquireAttribute("LagrangeVariant").Data()[0]
        if "Degree" not in adios_file.io.AvailableAttributes().keys():
            raise KeyError(f"Mesh degree not found in {filename}")
        degree = adios_file.io.InquireAttribute("Degree").Data()[0]

        if not legacy:
            time_name = "MeshTime"
            for i in range(adios_file.file.Steps()):
                if i > 0:
                    adios_file.file.BeginStep()
                if time_name in adios_file.io.AvailableVariables().keys():
                    arr = adios_file.io.InquireVariable(time_name)
                    time_shape = arr.Shape()
                    arr.SetSelection([[0], [time_shape[0]]])
                    times = np.empty(time_shape[0], dtype=adios_to_numpy_dtype[arr.Type()])
                    adios_file.file.Get(arr, times, adios2.Mode.Sync)
                    if times[0] == time:
                        break
                if i == adios_file.file.Steps() - 1:
                    raise KeyError(
                        f"No data associated with {time_name}={time} found in {filename}"
                    )

                adios_file.file.EndStep()

            if time_name not in adios_file.io.AvailableVariables().keys():
                raise KeyError(f"No data associated with {time_name}={time} found in {filename}")

        # Get mesh geometry
        if "Points" not in adios_file.io.AvailableVariables().keys():
            raise KeyError(f"Mesh coordinates not found at Points in {filename}")
        geometry = adios_file.io.InquireVariable("Points")
        x_shape = geometry.Shape()
        geometry_range = compute_local_range(comm, x_shape[0])
        geometry.SetSelection(
            [
                [geometry_range[0], 0],
                [geometry_range[1] - geometry_range[0], x_shape[1]],
            ]
        )
        mesh_geometry = np.empty(
            (geometry_range[1] - geometry_range[0], x_shape[1]),
            dtype=adios_to_numpy_dtype[geometry.Type()],
        )
        adios_file.file.Get(geometry, mesh_geometry, adios2.Mode.Deferred)
        adios_file.file.PerformGets()
        adios_file.file.EndStep()

    if read_from_partition:
        partition_graph = read_adjacency_list(
            adios,
            comm,
            filename,
            "PartitioningData",
            "PartitioningOffset",
            backend_args["engine"],
        )
    else:
        partition_graph = None

    return ReadMeshData(
        cells=mesh_topology,
        cell_type=cell_type,
        x=mesh_geometry,
        degree=degree,
        lvar=lvar,
        partition_graph=partition_graph,
    )


def write_meshtags(
    filename: str | Path,
    comm: MPI.Intracomm,
    data: MeshTagsData,
    backend_args: dict[str, Any] | None = None,
):
    """Write mesh tags to file.

    Args:
        filename: Path to file to write to
        comm: MPI communicator used in storage
        data: Internal data structure for the mesh tags to save to file
        backend_args: Arguments to backend
    """
    backend_args = {} if backend_args is None else backend_args
    io_name = backend_args.get("io_name", "MeshTagWriter")
    engine = backend_args.get("engine", "BP4")
    adios = adios2.ADIOS(comm)
    with ADIOSFile(
        adios=adios,
        filename=filename,
        mode=adios2.Mode.Append,
        engine=engine,
        io_name=io_name,
    ) as adios_file:
        adios_file.file.BeginStep()

        # Write meshtag topology
        topology_var = adios_file.io.DefineVariable(
            data.name + "_topology",
            data.indices,
            shape=[data.num_entities_global, data.num_dofs_per_entity],
            start=[data.local_start, 0],
            count=[len(data.indices), data.num_dofs_per_entity],
        )
        adios_file.file.Put(topology_var, data.indices, adios2.Mode.Sync)

        # Write meshtag values
        values_var = adios_file.io.DefineVariable(
            data.name + "_values",
            data.values,
            shape=[data.num_entities_global],
            start=[data.local_start],
            count=[len(data.indices)],
        )
        adios_file.file.Put(values_var, data.values, adios2.Mode.Sync)

        # Write meshtag dim
        adios_file.io.DefineAttribute(data.name + "_dim", np.array([data.dim], dtype=np.uint8))
        adios_file.file.PerformPuts()
        adios_file.file.EndStep()


def read_meshtags_data(
    filename: str | Path, comm: MPI.Intracomm, name: str, backend_args: dict[str, Any] | None = None
) -> MeshTagsData:
    """Read mesh tags from file.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the mesh tags to read
        backend_args: Arguments to backend

    Returns:
        Internal data structure for the mesh tags read from file
    """

    adios = adios2.ADIOS(comm)
    backend_args = get_default_backend_args(backend_args)
    io_name = backend_args.get("io_name", "MeshTagsReader")
    engine = backend_args["engine"]
    legacy = backend_args["legacy"]
    with ADIOSFile(
        adios=adios,
        filename=filename,
        mode=adios2.Mode.Read,
        engine=engine,
        io_name=io_name,
    ) as adios_file:
        if not legacy:
            # Get mesh cell type
            dim_attr_name = f"{name}_dim"
            step = 0
            for i in range(adios_file.file.Steps()):
                adios_file.file.BeginStep()
                if dim_attr_name in adios_file.io.AvailableAttributes().keys():
                    step = i
                    break
                adios_file.file.EndStep()
            if dim_attr_name not in adios_file.io.AvailableAttributes().keys():
                raise KeyError(f"{dim_attr_name} not found in {filename}")

            m_dim = adios_file.io.InquireAttribute(dim_attr_name)
            dim = int(m_dim.Data()[0])

            # Get mesh tags entites
            topology_name = f"{name}_topology"
            for i in range(step, adios_file.file.Steps()):
                if i > step:
                    adios_file.file.BeginStep()
                if topology_name in adios_file.io.AvailableVariables().keys():
                    break
                adios_file.file.EndStep()
            if topology_name not in adios_file.io.AvailableVariables().keys():
                raise KeyError(f"{topology_name} not found in {filename}")

            topology = adios_file.io.InquireVariable(topology_name)
            top_shape = topology.Shape()
            topology_range = compute_local_range(comm, top_shape[0])

            topology.SetSelection(
                [
                    [topology_range[0], 0],
                    [topology_range[1] - topology_range[0], top_shape[1]],
                ]
            )
            mesh_entities = np.empty(
                (topology_range[1] - topology_range[0], top_shape[1]), dtype=np.int64
            )
            adios_file.file.Get(topology, mesh_entities, adios2.Mode.Deferred)

            # Get mesh tags values
            values_name = f"{name}_values"
            if values_name not in adios_file.io.AvailableVariables().keys():
                raise KeyError(f"{values_name} not found")

            values = adios_file.io.InquireVariable(values_name)
            val_shape = values.Shape()
            assert val_shape[0] == top_shape[0]
            values.SetSelection([[topology_range[0]], [topology_range[1] - topology_range[0]]])
            tag_values = np.empty(
                (topology_range[1] - topology_range[0]), dtype=values.Type().strip("_t")
            )
            adios_file.file.Get(values, tag_values, adios2.Mode.Deferred)

            adios_file.file.PerformGets()
            adios_file.file.EndStep()
        else:
            # Get mesh cell type
            dim_attr_name = f"{name}_dim"
            assert adios_file.file.Steps() == 0
            if (ct_key := f"/{name}/topology/celltype") in adios_file.io.AvailableAttributes():
                cell_type = adios_file.io.InquireAttribute(ct_key)
            else:
                raise ValueError(f"Celltype not found for meshtags {name} in {filename}.")
            dim = dolfinx.mesh.cell_dim(dolfinx.mesh.to_type(cell_type.DataString()[0]))

            # Get mesh tags entites
            if (top_key := f"/{name}/topology") in adios_file.io.AvailableVariables():
                topology = adios_file.io.InquireVariable(top_key)
            else:
                raise ValueError(f"Topology not found for meshtags {name} in {filename}.")

            top_shape = topology.Shape()
            topology_range = compute_local_range(comm, top_shape[0])

            topology.SetSelection(
                [
                    [topology_range[0], 0],
                    [topology_range[1] - topology_range[0], top_shape[1]],
                ]
            )
            mesh_entities = np.empty(
                (topology_range[1] - topology_range[0], top_shape[1]), dtype=np.int64
            )
            adios_file.file.Get(topology, mesh_entities, adios2.Mode.Deferred)

            # Get mesh tags values
            if (val_key := f"/{name}/values") in adios_file.io.AvailableVariables():
                values = adios_file.io.InquireVariable(val_key)
            else:
                raise ValueError(f"Values not found for meshtags {name} in {filename}.")

            val_shape = values.Shape()
            assert val_shape[0] == top_shape[0]

            values.SetSelection([[topology_range[0]], [topology_range[1] - topology_range[0]]])
            tag_values = np.empty(
                (topology_range[1] - topology_range[0]), dtype=values.Type().strip("_t")
            )
            adios_file.file.Get(values, tag_values, adios2.Mode.Deferred)

            adios_file.file.PerformGets()
            adios_file.file.EndStep()

        return MeshTagsData(
            name=name, values=tag_values.astype(np.int32), indices=mesh_entities, dim=dim
        )


def read_dofmap(
    filename: str | Path, comm: MPI.Intracomm, name: str, backend_args: dict[str, Any] | None = None
) -> dolfinx.graph.AdjacencyList:
    """Read the dofmap of a function with a given name.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the function to read the dofmap for
        backend_args: Arguments to backend

    Returns:
        Dofmap as an AdjacencyList
    """
    backend_args = {} if backend_args is None else backend_args

    # Handles legacy io4dolfinx files, modern files, and custom location of dofmap.
    legacy = backend_args.get("legacy", False)
    xdofmap_path: str | None
    dofmap_path: str | None
    if (dofmap_path := backend_args.get("dofmap", None)) is None:
        if legacy:
            dofmap_path = "Dofmap"
        else:
            dofmap_path = f"{name}_dofmap"

    if (xdofmap_path := backend_args.get("offsets", None)) is None:
        if legacy:
            xdofmap_path = "XDofmap"
        else:
            xdofmap_path = f"{name}_XDofmap"

    engine = backend_args.get("engine", "BP4")

    adios = adios2.ADIOS(comm)
    check_file_exists(filename)
    assert isinstance(xdofmap_path, str)
    return read_adjacency_list(adios, comm, filename, dofmap_path, xdofmap_path, engine=engine)


def read_dofs(
    filename: str | Path,
    comm: MPI.Intracomm,
    name: str,
    time: float,
    backend_args: dict[str, Any] | None = None,
) -> tuple[npt.NDArray[np.float32 | np.float64 | np.complex64 | np.complex128], int]:
    """Read the dofs (values) of a function with a given name from a given timestep.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the function to read the dofs for
        time: Time stamp associated with the function to read
        backend_args: Arguments to backend

    Returns:
        Contiguous sequence of degrees of freedom (with respect to input data)
        and the global starting point on the process.
        Process 0 has [0, M), process 1 [M, N), process 2 [N, O) etc.
    """

    backend_args = {} if backend_args is None else backend_args
    legacy = backend_args.get("legacy", False)
    engine = backend_args.get("engine", "BP4")
    io_name = backend_args.get("io_name", f"{name}_FunctionReader")
    # Check that file contains the function to read
    adios = adios2.ADIOS(comm)
    check_file_exists(filename)

    if not legacy:
        with ADIOSFile(
            adios=adios,
            filename=filename,
            mode=adios2.Mode.Read,
            engine=engine,
            io_name=io_name,
        ) as adios_file:
            variables = set(
                sorted(
                    map(
                        lambda x: x.split("_time")[0],
                        filter(lambda x: x.endswith("_time"), adios_file.io.AvailableVariables()),
                    )
                )
            )
            if name not in variables:
                raise KeyError(f"{name} not found in {filename}. Did you mean one of {variables}?")

    if legacy:
        array_path = "Values"
    else:
        array_path = f"{name}_values"

    time_name = f"{name}_time"
    return read_array(adios, filename, array_path, engine, comm, time, time_name, legacy=legacy)


def read_cell_perms(
    comm: MPI.Intracomm, filename: Path | str, backend_args: dict[str, Any] | None = None
) -> npt.NDArray[np.uint32]:
    """
    Read cell permutation from file with given communicator,
    Split in continuous chunks based on number of cells in the mesh (global).

    Args:
        adios: The ADIOS instance
        comm: The MPI communicator used to read the data
        filename: Path to input file
        variable: Name of cell-permutation variable
        num_cells_global: Number of cells in the mesh (global)
        engine: Type of ADIOS engine to use for reading data

    Returns:
        Cell-permutations local to the process

    .. note::
        No MPI communication is done during this call
    """
    adios = adios2.ADIOS(comm)
    check_file_exists(filename)

    # Open ADIOS engine
    backend_args = {} if backend_args is None else backend_args
    engine = backend_args.get("engine", "BP4")

    cell_perms, _ = read_array(
        adios, filename, "CellPermutations", engine=engine, comm=comm, legacy=True
    )

    return cell_perms.astype(np.uint32)


def read_hdf5_array(
    comm: MPI.Intracomm,
    filename: Path | str,
    group: str,
    backend_args: dict[str, Any] | None = None,
) -> tuple[np.ndarray, int]:
    adios = adios2.ADIOS(comm)
    """Read an array from an HDF5 file.

    Args:
        comm: MPI communicator used in storage
        filename: Path to file to read from
        group: Group in HDF5 file where array is stored
        backend_args: Arguments to backend

    Returns:
        Tuple containing:
            - Numpy array read from file
            - Global starting point on the process.
                Process 0 has [0, M), process 1 [M, N), process 2 [N, O) etc.
    """
    return read_array(adios, filename, group, engine="HDF5", comm=comm, legacy=True)


def write_function(
    filename: Path,
    comm: MPI.Intracomm,
    u: FunctionData,
    time: float = 0.0,
    mode: FileMode = FileMode.append,
    backend_args: dict[str, Any] | None = None,
):
    """
    Write a function to file using ADIOS2

    Args:
        comm: MPI communicator used in storage
        u: Internal data structure for the function data to save to file
        filename: Path to file to write to
        engine: ADIOS2 engine to use
        mode: ADIOS2 mode to use (write or append)
        time: Time stamp associated with function
        io_name: Internal name used for the ADIOS IO object
    """
    adios_mode = convert_file_mode(mode)
    backend_args = get_default_backend_args(backend_args)
    engine = backend_args["engine"]
    io_name = backend_args.get("io_name", "{name}_writer")

    adios = adios2.ADIOS(comm)
    cell_permutations_exists = False
    dofmap_exists = False
    XDofmap_exists = False
    if mode == adios2.Mode.Append:
        cell_permutations_exists = check_variable_exists(
            adios, filename, "CellPermutations", engine=engine
        )
        dofmap_exists = check_variable_exists(adios, filename, f"{u.name}_dofmap", engine=engine)
        XDofmap_exists = check_variable_exists(adios, filename, f"{u.name}_XDofmap", engine=engine)

    with ADIOSFile(
        adios=adios, filename=filename, mode=adios_mode, engine=engine, io_name=io_name, comm=comm
    ) as adios_file:
        adios_file.file.BeginStep()

        if not cell_permutations_exists:
            # Add mesh permutations
            pvar = adios_file.io.DefineVariable(
                "CellPermutations",
                u.cell_permutations,
                shape=[u.num_cells_global],
                start=[u.local_cell_range[0]],
                count=[u.local_cell_range[1] - u.local_cell_range[0]],
            )
            adios_file.file.Put(pvar, u.cell_permutations)

        if not dofmap_exists:
            # Add dofmap
            dofmap_var = adios_file.io.DefineVariable(
                f"{u.name}_dofmap",
                u.dofmap_array,
                shape=[u.global_dofs_in_dofmap],
                start=[u.dofmap_range[0]],
                count=[u.dofmap_range[1] - u.dofmap_range[0]],
            )
            adios_file.file.Put(dofmap_var, u.dofmap_array)

        if not XDofmap_exists:
            # Add XDofmap
            xdofmap_var = adios_file.io.DefineVariable(
                f"{u.name}_XDofmap",
                u.dofmap_offsets,
                shape=[u.num_cells_global + 1],
                start=[u.local_cell_range[0]],
                count=[u.local_cell_range[1] - u.local_cell_range[0] + 1],
            )
            adios_file.file.Put(xdofmap_var, u.dofmap_offsets)

        val_var = adios_file.io.DefineVariable(
            f"{u.name}_values",
            u.values,
            shape=[u.num_dofs_global],
            start=[u.dof_range[0]],
            count=[u.dof_range[1] - u.dof_range[0]],
        )
        adios_file.file.Put(val_var, u.values)

        # Add time step to file
        t_arr = np.array([time], dtype=np.float64)
        time_var = adios_file.io.DefineVariable(
            f"{u.name}_time",
            t_arr,
            shape=[1],
            start=[0],
            count=[1 if comm.rank == 0 else 0],
        )
        adios_file.file.Put(time_var, t_arr)
        adios_file.file.PerformPuts()
        adios_file.file.EndStep()


def read_legacy_mesh(
    filename: Path | str, comm: MPI.Intracomm, group: str
) -> tuple[npt.NDArray[np.int64], npt.NDArray[np.floating], str | None]:
    """Read in the mesh topology, geometry and (optionally) cell type from a
    legacy DOLFIN HDF5-file.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        group: Group in HDF5 file where mesh is stored

    Returns:
        Tuple containing:
            - Topology as a (num_cells, num_vertices_per_cell) array of global vertex indices
            - Geometry as a (num_vertices, geometric_dimension) array of vertex coordinates
            - Cell type as a string (e.g. "tetrahedron") or None if not found
    """
    # Create ADIOS2 reader
    adios = adios2.ADIOS(comm)
    with ADIOSFile(
        adios=adios,
        filename=filename,
        mode=adios2.Mode.Read,
        io_name="Mesh reader",
        engine="HDF5",
    ) as adios_file:
        # Get mesh topology (distributed)
        if f"{group}/topology" not in adios_file.io.AvailableVariables().keys():
            raise KeyError(f"Mesh topology not found at '{group}/topology'")
        topology = adios_file.io.InquireVariable(f"{group}/topology")
        shape = topology.Shape()
        local_range = compute_local_range(comm, shape[0])
        topology.SetSelection([[local_range[0], 0], [local_range[1] - local_range[0], shape[1]]])

        mesh_topology = np.empty(
            (local_range[1] - local_range[0], shape[1]),
            dtype=topology.Type().strip("_t"),
        )
        adios_file.file.Get(topology, mesh_topology, adios2.Mode.Sync)

        # Get mesh cell type
        cell_type = None
        if f"{group}/topology/celltype" in adios_file.io.AvailableAttributes().keys():
            celltype = adios_file.io.InquireAttribute(f"{group}/topology/celltype")
            cell_type = celltype.DataString()[0]

        # Get mesh geometry

        for geometry_key in [f"{group}/geometry", f"{group}/coordinates"]:
            if geometry_key in adios_file.io.AvailableVariables().keys():
                break
        else:
            raise KeyError(
                f"Mesh coordinates not found at '{group}/coordinates' or '{group}/geometry'"
            )

        geometry = adios_file.io.InquireVariable(geometry_key)
        shape = geometry.Shape()
        local_range = compute_local_range(comm, shape[0])
        geometry.SetSelection([[local_range[0], 0], [local_range[1] - local_range[0], shape[1]]])
        mesh_geometry = np.empty(
            (local_range[1] - local_range[0], shape[1]),
            dtype=adios_to_numpy_dtype[geometry.Type()],
        )
        adios_file.file.Get(geometry, mesh_geometry, adios2.Mode.Sync)

    return mesh_topology, mesh_geometry, cell_type


def snapshot_checkpoint(
    filename: Path | str,
    mode: FileMode,
    u: dolfinx.fem.Function,
    backend_args: dict[str, Any] | None,
):
    """Create a snapshot checkpoint of a dolfinx function.

    Args:
        filename: Path to file to read from
        mode: File-mode to store the function
        u: dolfinx function to create a snapshot checkpoint for
        backend_args: Arguments to backend
    """
    adios_mode = convert_file_mode(mode)
    adios = adios2.ADIOS(u.function_space.mesh.comm)
    backend_args = {} if backend_args is None else backend_args
    io_name = backend_args.get("io_name", "SnapshotCheckPoint")
    engine = backend_args.get("engine", "BP4")
    with ADIOSFile(
        adios=adios,
        filename=filename,
        mode=adios_mode,
        io_name=io_name,
        engine=engine,
    ) as adios_file:
        if adios_mode == adios2.Mode.Write:
            dofmap = u.function_space.dofmap
            num_dofs_local = dofmap.index_map.size_local * dofmap.index_map_bs
            local_dofs = u.x.array[:num_dofs_local].copy()

            # Write to file
            adios_file.file.BeginStep()
            dofs = adios_file.io.DefineVariable("dofs", local_dofs, count=[num_dofs_local])
            adios_file.file.Put(dofs, local_dofs, adios2.Mode.Sync)
            adios_file.file.EndStep()
        elif adios_mode == adios2.Mode.Read:
            adios_file.file.BeginStep()
            in_variable = adios_file.io.InquireVariable("dofs")
            in_variable.SetBlockSelection(u.function_space.mesh.comm.rank)
            adios_file.file.Get(in_variable, u.x.array, adios2.Mode.Sync)
            adios_file.file.EndStep()
            u.x.scatter_forward()
        else:
            raise NotImplementedError(f"Mode {mode} is not implemented for snapshot checkpoint")


def read_function_names(
    filename: Path | str, comm: MPI.Intracomm, backend_args: dict[str, Any] | None
) -> list[str]:
    """Read all function names from a file.

    Args:
        filename: Path to file
        comm: MPI communicator to launch IO on.
        backend_args: Arguments to backend

    Returns:
        A list of function names.
    """
    raise NotImplementedError("Reading function names are not implemented with ADIOS2")


def read_point_data(
    filename: Path | str,
    name: str,
    comm: MPI.Intracomm,
    time: float | str | None,
    backend_args: dict[str, Any] | None,
) -> tuple[np.ndarray, int]:
    """Read data from the nodes of a mesh.

    Args:
        filename: Path to file
        name: Name of point data
        comm: Communicator to launch IO on.
        time: The time stamp
        backend_args: The backend arguments

    Returns:
       Data local to process (contiguous, no mpi comm) and local start range
    """
    raise NotImplementedError("The ADIOS2 backend cannot read point data.")


def read_cell_data(
    filename: Path | str,
    name: str,
    comm: MPI.Intracomm,
    time: str | float | None,
    backend_args: dict[str, Any] | None,
) -> tuple[npt.NDArray[np.int64], np.ndarray]:
    """Read data from the cells of a mesh.

    Args:
        filename: Path to file
        name: Name of point data
        comm: Communicator to launch IO on.
        time: The time stamp
        backend_args: The backend arguments
    Returns:
        A tuple (topology, dofs) where topology contains the
        vertex indices of the cells, dofs the degrees of
        freedom within that cell.
    """
    raise NotImplementedError("The ADIOS2 backend does not support reading cell data.")


def write_data(
    filename: Path | str,
    array_data: ArrayData,
    comm: MPI.Intracomm,
    time: str | float | None,
    mode: FileMode,
    backend_args: dict[str, Any] | None,
):
    """Write a 2D-array to file (distributed across proceses with MPI).


    Args:
        filename: Path to file
        array_data: Data to write to file
        comm: MPI communicator to open the file with.
        time: Time stamp
        mode: Append or write
        backend_args: The backend arguments
    """
    raise NotImplementedError("ADIOS2 has not implemented this yet")



"""
Module that uses pyvista to import grids.
"""

from typing import Any

import numpy as np
import numpy.typing as npt

try:
    import pyvista
except ImportError:
    raise ModuleNotFoundError("The PyVista-backend requires pyvista")
from pathlib import Path

from mpi4py import MPI

import basix
import dolfinx

from io4dolfinx.structures import ArrayData, FunctionData, MeshData, MeshTagsData, ReadMeshData
from io4dolfinx.utils import check_file_exists

from .. import FileMode, ReadMode

# Cell types can be found at
# https://vtk.org/doc/nightly/html/vtkCellType_8h_source.html
_first_order_vtk = {
    1: "point",
    3: "interval",
    5: "triangle",
    9: "quadrilateral",
    10: "tetrahedron",
    12: "hexahedron",
}

_arbitrary_lagrange_vtk = {
    68: "interval",
    69: "triangle",
    70: "quadrilateral",
    71: "tetrahedron",
    72: "hexahedron",
    73: "prism",
    74: "pyramid",
}


read_mode = ReadMode.serial


def _cell_degree(ct: str, num_nodes: int) -> int:
    if ct == "point":
        return 1
    elif ct == "interval":
        return int(num_nodes - 1)
    elif ct == "triangle":
        n = (np.sqrt(1 + 8 * num_nodes) - 1) / 2
        if 2 * num_nodes != n * (n + 1):
            raise ValueError(f"Unknown triangle layout. Number of nodes: {num_nodes}")
        return int(n - 1)
    elif ct == "tetrahedron":
        n = 0
        while n * (n + 1) * (n + 2) < 6 * num_nodes:
            n += 1
        if n * (n + 1) * (n + 2) != 6 * num_nodes:
            raise ValueError(f"Unknown tetrahedron layout. Number of nodes: {num_nodes}")
        return int(n - 1)

    elif ct == "quadrilateral":
        n = np.sqrt(num_nodes)
        if num_nodes != n * n:
            raise ValueError(f"Unknown quadrilateral layout. Number of nodes: {num_nodes}")
        return int(n - 1)
    elif ct == "hexahedron":
        n = np.cbrt(num_nodes)
        if num_nodes != n * n * n:
            raise ValueError(f"Unknown hexahedron layout. Number of nodes: {num_nodes}")
        return int(n - 1)
    elif ct == "prism":
        if num_nodes == 6:
            return 1
        elif num_nodes == 15:
            return 2
        else:
            raise ValueError(f"Unknown prism layout. Number of nodes: {num_nodes}")
    elif ct == "pyramid":
        if num_nodes == 5:
            return 1
        elif num_nodes == 13:
            return 2
        else:
            raise ValueError(f"Unknown pyramid layout. Number of nodes: {num_nodes}")
    else:
        raise ValueError(f"Unknown cell type {ct} with {num_nodes=}.")


def get_default_backend_args(arguments: dict[str, Any] | None) -> dict[str, Any]:
    """Get default backend arguments given a set of input arguments.

    Args:
        arguments: Input backend arguments

    Returns:
        Updated backend arguments
    """
    args = arguments or {}
    return args


def read_mesh_data(
    filename: Path | str,
    comm: MPI.Intracomm,
    time: str | float | None = None,
    read_from_partition: bool = False,
    backend_args: dict[str, Any] | None = None,
) -> ReadMeshData:
    """Read mesh data from file.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        time: Time stamp associated with the mesh to read
        read_from_partition: Whether to read partition information
        backend_args: Arguments to backend

    Returns:
        Internal data structure for the mesh data read from file
    """
    backend_args = get_default_backend_args(backend_args)
    check_file_exists(filename)
    if read_from_partition:
        raise RuntimeError("Cannot read partition data with Pyvista")
    cells: npt.NDArray[np.int64]
    geom: npt.NDArray[np.float64 | np.float32]
    if comm.rank == 0:
        in_data = pyvista.read(filename)
        if isinstance(in_data, pyvista.UnstructuredGrid):
            grid = in_data
        elif isinstance(in_data, pyvista.core.composite.MultiBlock):
            # To handle multiblock like pvd
            if hasattr(pyvista, "_VTK_SNAKE_CASE_STATE"):
                pyvista._VTK_SNAKE_CASE_STATE = "allow"
            else:
                # Compatibility with 0.47
                pyvista.core.vtk_snake_case._state = "allow"
            number_of_blocks = in_data.number_of_blocks
            assert number_of_blocks == 1
            b0 = in_data.get_block(0)
            assert isinstance(b0, pyvista.UnstructuredGrid)
            grid = b0
        else:
            raise RuntimeError(f"Unknown data type {type(in_data)}")
        geom = grid.points
        num_cells_global = grid.number_of_cells
        cells = grid.cells.reshape(num_cells_global, -1).astype(np.int64)
        nodes_per_cell_type = cells[:, 0]
        assert np.allclose(nodes_per_cell_type, nodes_per_cell_type[0]), "Single celltype support"
        cells = cells[:, 1:].astype(np.int64)
        cell_types = grid.celltypes
        assert len(np.unique(cell_types)) == 1
        if (cell_type := cell_types[0]) in _first_order_vtk.keys():
            cell_type = _first_order_vtk[cell_type]
            order = 1
        elif cell_type in _arbitrary_lagrange_vtk.keys():
            cell_type = _arbitrary_lagrange_vtk[cell_type]
            order = _cell_degree(cell_type, cells.shape[1])
        perm = dolfinx.cpp.io.perm_vtk(dolfinx.mesh.to_type(cell_type), cells.shape[1])
        cells = cells[:, perm]
        lvar = int(basix.LagrangeVariant.equispaced)
        gtype = backend_args.get("dtype", geom.dtype)
        order, lvar, nodes_per_cell, cell_type, gtype, gdim = comm.bcast(
            (order, lvar, cells.shape[1], cell_type, gtype, geom.shape[1]), root=0
        )
    else:
        order, lvar, nodes_per_cell, cell_type, gtype, gdim = comm.bcast(None, root=0)
        geom = np.zeros((0, gdim), dtype=gtype)
        cells = np.zeros((0, nodes_per_cell), dtype=np.int64)

    return ReadMeshData(
        cells=cells, cell_type=cell_type, x=geom.astype(gtype), lvar=lvar, degree=order
    )


def read_point_data(
    filename: Path | str,
    name: str,
    comm: MPI.Intracomm,
    time: float | str | None,
    backend_args: dict[str, Any] | None,
) -> tuple[np.ndarray, int]:
    """Read data from the nodes of a mesh.

    Args:
        filename: Path to file
        name: Name of point data
        comm: Communicator to launch IO on.
        time: The time stamp
        backend_args: The backend arguments

    Returns:
       Data local to process (contiguous, no mpi comm) and local start range
    """
    dataset: np.ndarray
    if MPI.COMM_WORLD.rank == 0:
        in_data = pyvista.read(filename)
        if isinstance(in_data, pyvista.UnstructuredGrid):
            grid = in_data
        elif isinstance(in_data, pyvista.core.composite.MultiBlock):
            # To handle multiblock like pvd
            if hasattr(pyvista, "_VTK_SNAKE_CASE_STATE"):
                pyvista._VTK_SNAKE_CASE_STATE = "allow"
            else:
                # Compatibility with 0.47
                pyvista.core.vtk_snake_case._state = "allow"
            number_of_blocks = in_data.number_of_blocks
            assert number_of_blocks == 1
            b0 = in_data.get_block(0)
            assert isinstance(b0, pyvista.UnstructuredGrid)
            grid = b0

        dataset = grid.point_data[name]
        if len(dataset.shape) == 1:
            num_components = 1
            dataset = dataset.reshape(-1, num_components)
        else:
            num_components = dataset.shape[1]
        if np.issubdtype(dataset.dtype, np.integer):
            gtype = grid.points.dtype
            dataset = dataset.astype(gtype)
        else:
            gtype = dataset.dtype
        num_components, gtype = comm.bcast((num_components, gtype), root=0)
        local_range_start = 0
    else:
        num_components, gtype = comm.bcast(None, root=0)
        dataset = np.zeros((0, num_components), dtype=gtype)
        local_range_start = 0

    return dataset, int(local_range_start)


def read_cell_data(
    filename: Path | str,
    name: str,
    comm: MPI.Intracomm,
    time: str | float | None,
    backend_args: dict[str, Any] | None,
) -> tuple[npt.NDArray[np.int64], np.ndarray]:
    dataset: np.ndarray
    topology: np.ndarray
    if MPI.COMM_WORLD.rank == 0:
        in_data = pyvista.read(filename)
        if isinstance(in_data, pyvista.UnstructuredGrid):
            grid = in_data
        elif isinstance(in_data, pyvista.core.composite.MultiBlock):
            # To handle multiblock like pvd
            if hasattr(pyvista, "_VTK_SNAKE_CASE_STATE"):
                pyvista._VTK_SNAKE_CASE_STATE = "allow"
            else:
                # Compatibility with 0.47
                pyvista.core.vtk_snake_case._state = "allow"
            number_of_blocks = in_data.number_of_blocks
            assert number_of_blocks == 1
            b0 = in_data.get_block(0)
            assert isinstance(b0, pyvista.UnstructuredGrid)
            grid = b0

        dataset = grid.cell_data[name]
        if len(dataset.shape) == 1:
            num_components = 1
            dataset = dataset.reshape(-1, num_components)
        else:
            num_components = dataset.shape[1]

        if np.issubdtype(dataset.dtype, np.integer):
            gtype = in_data.points.dtype
            dataset = dataset.astype(gtype)
        else:
            gtype = dataset.dtype
        num_components, gtype = comm.bcast((num_components, gtype), root=0)
    else:
        num_components, gtype = comm.bcast(None, root=0)
        dataset = np.zeros((0, num_components), dtype=gtype)
    _time = float(time) if time is not None else None
    topology = read_mesh_data(filename, comm, _time, False, backend_args=None).cells
    return topology, dataset


def write_attributes(
    filename: Path | str,
    comm: MPI.Intracomm,
    name: str,
    attributes: dict[str, np.ndarray],
    backend_args: dict[str, Any] | None,
):
    """Write attributes to file.

    Args:
        filename: Path to file to write to
        comm: MPI communicator used in storage
        name: Name of the attribute group
        attributes: Dictionary of attributes to write
        backend_args: Arguments to backend
    """
    raise NotImplementedError("The Pyvista backend cannot write attributes.")


def read_attributes(
    filename: Path | str,
    comm: MPI.Intracomm,
    name: str,
    backend_args: dict[str, Any] | None,
) -> dict[str, Any]:
    """Read attributes from file.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the attribute group
        backend_args: Arguments to backend

    Returns:
        Dictionary of attributes read from file
    """
    raise NotImplementedError("The Pyvista backend cannot read attributes.")


def read_timestamps(
    filename: Path | str,
    comm: MPI.Intracomm,
    function_name: str,
    backend_args: dict[str, Any] | None,
) -> npt.NDArray[np.float64 | str]:  # type: ignore[type-var]
    """Read timestamps from file.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        function_name: Name of the function to read timestamps for
        backend_args: Arguments to backend

    Returns:
        Numpy array of timestamps read from file
    """
    raise NotImplementedError("The Pyvista backend cannot read timestamps.")


def read_function_names(
    filename: Path | str, comm: MPI.Intracomm, backend_args: dict[str, Any] | None
) -> list[str]:
    """Read all function names from a file.

    Args:
        filename: Path to file
        comm: MPI communicator to launch IO on.
        backend_args: Arguments to backend

    Returns:
        A list of function names.
    """
    in_data = pyvista.read(filename)
    if isinstance(in_data, pyvista.UnstructuredGrid):
        grid = in_data
    elif isinstance(in_data, pyvista.core.composite.MultiBlock):
        # To handle multiblock like pvd
        if hasattr(pyvista, "_VTK_SNAKE_CASE_STATE"):
            pyvista._VTK_SNAKE_CASE_STATE = "allow"
        else:
            # Compatibility with 0.47
            pyvista.core.vtk_snake_case._state = "allow"
        number_of_blocks = in_data.number_of_blocks
        assert number_of_blocks == 1
        b0 = in_data.get_block(0)
        assert isinstance(b0, pyvista.UnstructuredGrid)
        grid = b0

    point_data = list(grid.point_data.keys())
    cell_data = list(grid.cell_data.keys())
    return point_data + cell_data


def write_mesh(
    filename: Path | str,
    comm: MPI.Intracomm,
    mesh: MeshData,
    backend_args: dict[str, Any] | None,
    mode: FileMode,
    time: float,
):
    """
    Write a mesh to file.

    Args:
        comm: MPI communicator used in storage
        mesh: Internal data structure for the mesh data to save to file
        filename: Path to file to write to
        backend_args: Arguments to backend
        mode: File-mode to store the mesh
        time: Time stamp associated with the mesh
    """
    raise NotImplementedError("The Pyvista backend cannot write meshes.")


def write_meshtags(
    filename: str | Path,
    comm: MPI.Intracomm,
    data: MeshTagsData,
    backend_args: dict[str, Any] | None,
):
    """Write mesh tags to file.

    Args:
        filename: Path to file to write to
        comm: MPI communicator used in storage
        data: Internal data structure for the mesh tags to save to file
        backend_args: Arguments to backend
    """
    raise NotImplementedError("The Pyvista backend cannot write meshtags.")


def read_meshtags_data(
    filename: str | Path,
    comm: MPI.Intracomm,
    name: str,
    backend_args: dict[str, Any] | None,
) -> MeshTagsData:
    """Read mesh tags from file.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the mesh tags to read
        backend_args: Arguments to backend

    Returns:
        Internal data structure for the mesh tags read from file
    """
    raise NotImplementedError("The Pyvista backend cannot read meshtags.")


def read_dofmap(
    filename: str | Path,
    comm: MPI.Intracomm,
    name: str,
    backend_args: dict[str, Any] | None,
) -> dolfinx.graph.AdjacencyList:
    """Read the dofmap of a function with a given name.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the function to read the dofmap for
        backend_args: Arguments to backend

    Returns:
        Dofmap as an AdjacencyList
    """
    raise NotImplementedError("The Pyvista backend cannot make checkpoints.")


def read_dofs(
    filename: str | Path,
    comm: MPI.Intracomm,
    name: str,
    time: float,
    backend_args: dict[str, Any] | None,
) -> tuple[npt.NDArray[np.float32 | np.float64 | np.complex64 | np.complex128], int]:
    """Read the dofs (values) of a function with a given name from a given timestep.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the function to read the dofs for
        time: Time stamp associated with the function to read
        backend_args: Arguments to backend

    Returns:
        Contiguous sequence of degrees of freedom (with respect to input data)
        and the global starting point on the process.
        Process 0 has [0, M), process 1 [M, N), process 2 [N, O) etc.
    """
    raise NotImplementedError("The Pyvista backend cannot make checkpoints.")


def read_cell_perms(
    comm: MPI.Intracomm, filename: Path | str, backend_args: dict[str, Any] | None
) -> npt.NDArray[np.uint32]:
    """
    Read cell permutation from file with given communicator,
    Split in continuous chunks based on number of cells in the input data.

    Args:
        comm: MPI communicator used in storage
        filename: Path to file to read from
        backend_args: Arguments to backend

    Returns:
        Contiguous sequence of permutations (with respect to input data)
        Process 0 has [0, M), process 1 [M, N), process 2 [N, O) etc.
    """
    raise NotImplementedError("The Pyvista backend cannot make checkpoints.")


def write_function(
    filename: Path,
    comm: MPI.Intracomm,
    u: FunctionData,
    time: float,
    mode: FileMode,
    backend_args: dict[str, Any] | None,
):
    """
    Write a function to file.

    Args:
        comm: MPI communicator used in storage
        u: Internal data structure for the function data to save to file
        filename: Path to file to write to
        time: Time stamp associated with function
        mode: File-mode to store the function
        backend_args: Arguments to backend
    """
    raise NotImplementedError("The Pyvista backend cannot make checkpoints.")


def read_legacy_mesh(
    filename: Path | str, comm: MPI.Intracomm, group: str
) -> tuple[npt.NDArray[np.int64], npt.NDArray[np.floating], str | None]:
    """Read in the mesh topology, geometry and (optionally) cell type from a
    legacy DOLFIN HDF5-file.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        group: Group in HDF5 file where mesh is stored

    Returns:
        Tuple containing:
            - Topology as a (num_cells, num_vertices_per_cell) array of global vertex indices
            - Geometry as a (num_vertices, geometric_dimension) array of vertex coordinates
            - Cell type as a string (e.g. "tetrahedron") or None if not found
    """
    raise NotImplementedError("The Pyvista backend cannot read legacy DOLFIN meshes.")


def snapshot_checkpoint(
    filename: Path | str,
    mode: FileMode,
    u: dolfinx.fem.Function,
    backend_args: dict[str, Any] | None,
):
    """Create a snapshot checkpoint of a dolfinx function.

    Args:
        filename: Path to file to read from
        mode: File-mode to store the function
        u: dolfinx function to create a snapshot checkpoint for
        backend_args: Arguments to backend
    """
    raise NotImplementedError("The Pyvista backend cannot make checkpoints.")


def read_hdf5_array(
    comm: MPI.Intracomm,
    filename: Path | str,
    group: str,
    backend_args: dict[str, Any] | None,
) -> tuple[np.ndarray, int]:
    """Read an array from an HDF5 file.

    Args:
        comm: MPI communicator used in storage
        filename: Path to file to read from
        group: Group in HDF5 file where array is stored
        backend_args: Arguments to backend

    Returns:
        Tuple containing:
            - Numpy array read from file
            - Global starting point on the process.
                Process 0 has [0, M), process 1 [M, N), process 2 [N, O) etc.
    """
    raise NotImplementedError("The Pyvista backend cannot read HDF5 arrays")


def write_data(
    filename: Path | str,
    array_data: ArrayData,
    comm: MPI.Intracomm,
    time: str | float | None,
    mode: FileMode,
    backend_args: dict[str, Any] | None,
):
    """Write a 2D-array to file (distributed across proceses with MPI).

    Args:
        filename: Path to file
        array_data: Data to write to file
        comm: MPI communicator to open the file with
        time: Time stamp
        mode: Append or write
        backend_args: The backend arguments
    """
    raise NotImplementedError("The pyvista backend does not support writing point data")



# Copyright (C) 2024 Jørgen Schartum Dokken
#
# This file is part of io4dolfinx.
#
# SPDX-License-Identifier:    MIT


"""
Functions to create checkpoints with adios4dolfinx v0.7.x
"""

import argparse
import pathlib
from importlib.metadata import version

from mpi4py import MPI

import adios4dolfinx
import dolfinx
import numpy as np

a4d_version = version("adios4dolfinx")
assert a4d_version < "0.7.2", (
    f"Creating a legacy checkpoint requires adios4dolfinx < 0.7.2, you have {a4d_version}."
)


def f(x):
    values = np.zeros((2, x.shape[1]), dtype=np.float64)
    values[0] = x[0]
    values[1] = -x[1]
    return values


def write_checkpoint(filename, mesh, el, f):
    V = dolfinx.fem.FunctionSpace(mesh, el)
    uh = dolfinx.fem.Function(V, dtype=np.float64)
    uh.interpolate(f)

    adios4dolfinx.write_mesh(V.mesh, filename)
    adios4dolfinx.write_function(uh, filename)


def verify_checkpoint(filename, el, f):
    mesh = adios4dolfinx.read_mesh(
        MPI.COMM_WORLD, filename, "BP4", dolfinx.mesh.GhostMode.shared_facet
    )
    V = dolfinx.fem.FunctionSpace(mesh, el)
    uh = dolfinx.fem.Function(V, dtype=np.float64)
    adios4dolfinx.read_function(uh, filename)

    u_ex = dolfinx.fem.Function(V, dtype=np.float64)
    u_ex.interpolate(f)

    np.testing.assert_allclose(u_ex.x.array, uh.x.array, atol=1e-15)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--output-dir", type=str, default="legacy_checkpoint", dest="dir")

    inputs = parser.parse_args()
    path = pathlib.Path(inputs.dir)
    path.mkdir(exist_ok=True, parents=True)
    filename = path / "adios4dolfinx_checkpoint.bp"

    mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 10, 10)
    el = ("N1curl", 3)
    write_checkpoint(filename, mesh, el, f)
    MPI.COMM_WORLD.Barrier()
    verify_checkpoint(filename, el, f)



# # Writing a function checkpoint
# In the previous sections, we have gone in to quite some detail as to how
# to store meshes with io4dolfinx.
# This section will explain how to store {py:class}`functions<dolfinx.fem.Function>`,
# and how to read them back in.

# We start by creating a {py:class}`mesh<dolfinx.mesh.Mesh>`

# +
import logging
from pathlib import Path

from mpi4py import MPI

import dolfinx
import ipyparallel as ipp

import io4dolfinx

assert MPI.COMM_WORLD.size == 1, "This example should only be run with 1 MPI process"

mesh = dolfinx.mesh.create_unit_square(
    MPI.COMM_WORLD, nx=10, ny=10, cell_type=dolfinx.cpp.mesh.CellType.quadrilateral
)
# -

# Next, we create a function, and interpolate a polynomial function into the function space

# +
el = "N1curl"
degree = 3
V = dolfinx.fem.functionspace(mesh, (el, degree))


def f(x):
    return -(x[1] ** 2), x[0] - 2 * x[1]


u = dolfinx.fem.Function(V)
u.interpolate(f)
# -

# For the checkpointing, we start by storing the mesh to file

filename = Path("function_checkpoint.bp")
io4dolfinx.write_mesh(filename, mesh)

# Next, we store the function to file, and associate it with a name.
# Note that we can also associate a time stamp with it, as done for meshes in
# [Writing time-dependent mesh checkpoint](./time_dependent_mesh).
# We use {py:func}`io4dolfinx.write_function` for this.

io4dolfinx.write_function(filename, u, time=0.3, name="my_curl_function")

# Next, we want to read the function back in (using multiple MPI processes)
# and check that the function is correct.
# We use {py:func}`io4dolfinx.read_function` for this.
#
# ```{admonition} What mesh to use?
# Note that we have read in the mesh with {py:func}`io4dolfinx.read_mesh`
# before reading in the function.
# We **cannot** use {py:func}`io4dolfinx.read_function` to read function data to the
# original `mesh` object.
# To do this, we need to use {py:func}`io4dolfinx.write_function_on_input_mesh`, see
# [Writing function on input mesh checkpoint](./original_checkpoint).
# for more details.
# ```


def read_function(filename: Path, timestamp: float):
    from mpi4py import MPI

    import dolfinx
    import numpy as np

    import io4dolfinx

    in_mesh = io4dolfinx.read_mesh(filename, MPI.COMM_WORLD)
    W = dolfinx.fem.functionspace(in_mesh, (el, degree))
    u_ref = dolfinx.fem.Function(W)
    u_ref.interpolate(f)
    u_in = dolfinx.fem.Function(W)
    io4dolfinx.read_function(filename, u_in, time=timestamp, name="my_curl_function")
    np.testing.assert_allclose(u_ref.x.array, u_in.x.array, atol=1e-14)
    print(
        f"{MPI.COMM_WORLD.rank + 1}/{MPI.COMM_WORLD.size}: ",
        f"Function read in correctly at time {timestamp}",
    )


with ipp.Cluster(engines="mpi", n=3, log_level=logging.ERROR) as cluster:
    cluster[:].push({"f": f, "el": el, "degree": degree})
    query = cluster[:].apply_async(read_function, filename, 0.3)
    query.wait()
    assert query.successful(), query.error
    print("".join(query.stdout))



# # Checkpoint on input mesh
# As we have discussed earlier, one can choose to store function data in a way that
# is N-to-M compatible by using {py:func}`io4dolfinx.write_checkpoint`.
# This stores the distributed mesh in it's current (partitioned) ordering, and does
# use the original input data ordering for the cells and connectivity.
# This means that you cannot use your original mesh (from `.xdmf` files) or mesh tags
# together with the checkpoint. The checkpoint has to store the mesh and associated
# mesh-tags.

# An optional way of store an N-to-M checkpoint is to store the function data in the same
# ordering as the mesh. The write operation will be more expensive, as it requires data
# communication to ensure contiguous data being written to the checkpoint.
# The method is exposed as {py:func}`io4dolfinx.write_function_on_input_mesh`.
# Below we will demonstrate this method.

# +
import logging
from pathlib import Path
from typing import Tuple

import ipyparallel as ipp


def locate_facets(x, tol=1.0e-12):
    return abs(x[0]) < tol


def create_xdmf_mesh(filename: Path):
    from mpi4py import MPI

    import dolfinx

    mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 10, 10)
    facets = dolfinx.mesh.locate_entities_boundary(mesh, mesh.topology.dim - 1, locate_facets)
    facet_tag = dolfinx.mesh.meshtags(mesh, mesh.topology.dim - 1, facets, 1)
    with dolfinx.io.XDMFFile(MPI.COMM_WORLD, filename.with_suffix(".xdmf"), "w") as xdmf:
        xdmf.write_mesh(mesh)
        xdmf.write_meshtags(facet_tag, mesh.geometry)
    print(f"{mesh.comm.rank + 1}/{mesh.comm.size} Mesh written to {filename.with_suffix('.xdmf')}")


mesh_file = Path("MyMesh.xdmf")
with ipp.Cluster(engines="mpi", n=4, log_level=logging.ERROR) as cluster:
    # Create a mesh and write to XDMFFile
    cluster[:].push({"locate_facets": locate_facets})
    query = cluster[:].apply_async(create_xdmf_mesh, mesh_file)
    query.wait()
    assert query.successful(), query.error
    print("".join(query.stdout))
# -

# Next, we will create a function on the mesh and write it to a checkpoint.


# +
def f(x):
    return (x[0] + x[1]) * (x[0] < 0.5), x[1], x[2] - x[1]


def write_function(
    mesh_filename: Path, function_filename: Path, element: Tuple[str, int, Tuple[int,]]
):
    from mpi4py import MPI

    import dolfinx

    import io4dolfinx

    with dolfinx.io.XDMFFile(MPI.COMM_WORLD, mesh_filename, "r") as xdmf:
        mesh = xdmf.read_mesh()
    V = dolfinx.fem.functionspace(mesh, element)
    u = dolfinx.fem.Function(V)
    u.interpolate(f)

    io4dolfinx.write_function_on_input_mesh(
        function_filename.with_suffix(".bp"),
        u,
        mode=io4dolfinx.FileMode.write,
        time=0.0,
        name="Output",
    )
    print(
        f"{mesh.comm.rank + 1}/{mesh.comm.size} Function written to ",
        f"{function_filename.with_suffix('.bp')}",
    )


# -

# Read in mesh and write function to file

element = ("DG", 4, (3,))
function_file = Path("MyFunction.bp")
with ipp.Cluster(engines="mpi", n=2, log_level=logging.ERROR) as cluster:
    cluster[:].push({"f": f})
    query = cluster[:].apply_async(write_function, mesh_file, function_file, element)
    query.wait()
    assert query.successful(), query.error
    print("".join(query.stdout))


# Finally, we will read in the mesh from file and the function from the checkpoint
# and compare it with the analytical solution.


def verify_checkpoint(
    mesh_filename: Path, function_filename: Path, element: Tuple[str, int, Tuple[int,]]
):
    from mpi4py import MPI

    import dolfinx
    import numpy as np

    import io4dolfinx

    with dolfinx.io.XDMFFile(MPI.COMM_WORLD, mesh_filename, "r") as xdmf:
        in_mesh = xdmf.read_mesh()
    V = dolfinx.fem.functionspace(in_mesh, element)
    u_in = dolfinx.fem.Function(V)
    io4dolfinx.read_function(function_filename.with_suffix(".bp"), u_in, time=0.0, name="Output")

    # Compute exact interpolation
    u_ex = dolfinx.fem.Function(V)
    u_ex.interpolate(f)

    np.testing.assert_allclose(u_in.x.array, u_ex.x.array)
    print(
        "Successfully read checkpoint onto mesh on rank ",
        f"{in_mesh.comm.rank + 1}/{in_mesh.comm.size}",
    )


# Verify checkpoint by comparing to exact solution

with ipp.Cluster(engines="mpi", n=5, log_level=logging.ERROR) as cluster:
    cluster[:].push({"f": f})
    query = cluster[:].apply_async(verify_checkpoint, mesh_file, function_file, element)
    query.wait()
    assert query.successful(), query.error
    print("".join(query.stdout))
